from __future__ import absolute_import, division, print_function, unicode_literals

from ctypes import LittleEndianStructure
from ctypes.wintypes import USHORT, ULONG

from cwinsdk.shared.ntdef import UCHAR

# Windows Kits\10\Include\10.0.17763.0\km\ata.h

class GeneralConfiguration(LittleEndianStructure):
	__slots__ = ()
	_pack_ = 1
	_fields_ = [
		("Reserved1", USHORT, 1),
		("Retired3", USHORT, 1),
		("ResponseIncomplete", USHORT, 1),
		("Retired2", USHORT, 3),
		("FixedDevice", USHORT, 1),
		("RemovableMedia", USHORT, 1),
		("Retired1", USHORT, 7),
		("DeviceType", USHORT, 1),
	]

class TrustedComputing(LittleEndianStructure):
	__slots__ = ()
	_pack_ = 1
	_fields_ = [
		("FeatureSupported", USHORT, 1),
		("Reserved", USHORT, 15),
	]

class Capabilities(LittleEndianStructure):
	__slots__ = ()
	_pack_ = 1
	_fields_ = [
		("CurrentLongPhysicalSectorAlignment", UCHAR, 2),
		("ReservedByte49", UCHAR, 6),
		("DmaSupported", UCHAR, 1),
		("LbaSupported", UCHAR, 1),
		("IordyDisable", UCHAR, 1),
		("IordySupported", UCHAR, 1),
		("Reserved1", UCHAR, 1),
		("StandybyTimerSupport", UCHAR, 1),
		("Reserved2", UCHAR, 2),
		("ReservedWord50", USHORT),
	]

class AdditionalSupported(LittleEndianStructure):
	__slots__ = ()
	_pack_ = 1
	_fields_ = [
		("ZonedCapabilities", USHORT, 2),
		("NonVolatileWriteCache", USHORT, 1),
		("ExtendedUserAddressableSectorsSupported", USHORT, 1),
		("DeviceEncryptsAllUserData", USHORT, 1),
		("ReadZeroAfterTrimSupported", USHORT, 1),
		("Optional28BitCommandsSupported", USHORT, 1),
		("IEEE1667", USHORT, 1),
		("DownloadMicrocodeDmaSupported", USHORT, 1),
		("SetMaxSetPasswordUnlockDmaSupported", USHORT, 1),
		("WriteBufferDmaSupported", USHORT, 1),
		("ReadBufferDmaSupported", USHORT, 1),
		("DeviceConfigIdentifySetDmaSupported", USHORT, 1),
		("LPSAERCSupported", USHORT, 1),
		("DeterministicReadAfterTrimSupported", USHORT, 1),
		("CFastSpecSupported", USHORT, 1),
	]

class SerialAtaCapabilities(LittleEndianStructure):
	__slots__ = ()
	_pack_ = 1
	_fields_ = [
		("Reserved0", USHORT, 1), # shall be set to 0
		("SataGen1", USHORT, 1),
		("SataGen2", USHORT, 1),
		("SataGen3", USHORT, 1),
		("Reserved1", USHORT, 4),
		("NCQ", USHORT, 1),
		("HIPM", USHORT, 1),
		("PhyEvents", USHORT, 1),
		("NcqUnload", USHORT, 1),
		("NcqPriority", USHORT, 1),
		("HostAutoPS", USHORT, 1),
		("DeviceAutoPS", USHORT, 1),
		("ReadLogDMA", USHORT, 1),

		("Reserved2", USHORT, 1), # shall be set to 0
		("CurrentSpeed", USHORT, 3),
		("NcqStreaming", USHORT, 1),
		("NcqQueueMgmt", USHORT, 1),
		("NcqReceiveSend", USHORT, 1),
		("DEVSLPtoReducedPwrState", USHORT, 1),
		("Reserved3", USHORT, 8),
	]

class SerialAtaFeaturesSupported(LittleEndianStructure):
	__slots__ = ()
	_pack_ = 1
	_fields_ = [
		("Reserved0", USHORT, 1), # shall be set to 0
		("NonZeroOffsets", USHORT, 1),
		("DmaSetupAutoActivate", USHORT, 1),
		("DIPM", USHORT, 1),
		("InOrderData", USHORT, 1),
		("HardwareFeatureControl", USHORT, 1),
		("SoftwareSettingsPreservation", USHORT, 1),
		("NCQAutosense", USHORT, 1),
		("DEVSLP", USHORT, 1),
		("HybridInformation", USHORT, 1),
		("Reserved1", USHORT, 6),
	]

class SerialAtaFeaturesEnabled(LittleEndianStructure):
	__slots__ = ()
	_pack_ = 1
	_fields_ = [
		("Reserved0", USHORT, 1), # shall be set to 0
		("NonZeroOffsets", USHORT, 1),
		("DmaSetupAutoActivate", USHORT, 1),
		("DIPM", USHORT, 1),
		("InOrderData", USHORT, 1),
		("HardwareFeatureControl", USHORT, 1),
		("SoftwareSettingsPreservation", USHORT, 1),
		("DeviceAutoPS", USHORT, 1),
		("DEVSLP", USHORT, 1),
		("HybridInformation", USHORT, 1),
		("Reserved1", USHORT, 6),
	]

class CommandSetSupport(LittleEndianStructure):
	__slots__ = ()
	_pack_ = 1
	_fields_ = [
		("SmartCommands", USHORT, 1),
		("SecurityMode", USHORT, 1),
		("RemovableMediaFeature", USHORT, 1),
		("PowerManagement", USHORT, 1), # shall be set to 1
		("Reserved1", USHORT, 1),
		("WriteCache", USHORT, 1),
		("LookAhead", USHORT, 1),
		("ReleaseInterrupt", USHORT, 1),
		("ServiceInterrupt", USHORT, 1),
		("DeviceReset", USHORT, 1),
		("HostProtectedArea", USHORT, 1),
		("Obsolete1", USHORT, 1),
		("WriteBuffer", USHORT, 1),
		("ReadBuffer", USHORT, 1),
		("Nop", USHORT, 1),
		("Obsolete2", USHORT, 1),

		("DownloadMicrocode", USHORT, 1),
		("DmaQueued", USHORT, 1),
		("Cfa", USHORT, 1),
		("AdvancedPm", USHORT, 1),
		("Msn", USHORT, 1),
		("PowerUpInStandby", USHORT, 1),
		("ManualPowerUp", USHORT, 1),
		("Reserved2", USHORT, 1),
		("SetMax", USHORT, 1),
		("Acoustics", USHORT, 1),
		("BigLba", USHORT, 1),
		("DeviceConfigOverlay", USHORT, 1),
		("FlushCache", USHORT, 1),
		("FlushCacheExt", USHORT, 1),
		("WordValid83", USHORT, 2), # shall be 01b

		("SmartErrorLog", USHORT, 1),
		("SmartSelfTest", USHORT, 1),
		("MediaSerialNumber", USHORT, 1),
		("MediaCardPassThrough", USHORT, 1),
		("StreamingFeature", USHORT, 1),
		("GpLogging", USHORT, 1),
		("WriteFua", USHORT, 1),
		("WriteQueuedFua", USHORT, 1),
		("WWN64Bit", USHORT, 1),
		("URGReadStream", USHORT, 1),
		("URGWriteStream", USHORT, 1),
		("ReservedForTechReport", USHORT, 2),
		("IdleWithUnloadFeature", USHORT, 1),
		("WordValid", USHORT, 2), # shall be 01b
	]

class CommandSetActive(LittleEndianStructure):
	__slots__ = ()
	_pack_ = 1
	_fields_ = [
		("SmartCommands", USHORT, 1),
		("SecurityMode", USHORT, 1),
		("RemovableMediaFeature", USHORT, 1),
		("PowerManagement", USHORT, 1),
		("Reserved1", USHORT, 1),
		("WriteCache", USHORT, 1),
		("LookAhead", USHORT, 1),
		("ReleaseInterrupt", USHORT, 1),
		("ServiceInterrupt", USHORT, 1),
		("DeviceReset", USHORT, 1),
		("HostProtectedArea", USHORT, 1),
		("Obsolete1", USHORT, 1),
		("WriteBuffer", USHORT, 1),
		("ReadBuffer", USHORT, 1),
		("Nop", USHORT, 1),
		("Obsolete2", USHORT, 1),

		("DownloadMicrocode", USHORT, 1),
		("DmaQueued", USHORT, 1),
		("Cfa", USHORT, 1),
		("AdvancedPm", USHORT, 1),
		("Msn", USHORT, 1),
		("PowerUpInStandby", USHORT, 1),
		("ManualPowerUp", USHORT, 1),
		("Reserved2", USHORT, 1),
		("SetMax", USHORT, 1),
		("Acoustics", USHORT, 1),
		("BigLba", USHORT, 1),
		("DeviceConfigOverlay", USHORT, 1),
		("FlushCache", USHORT, 1),
		("FlushCacheExt", USHORT, 1),
		("Resrved3", USHORT, 1),
		("Words119_120Valid", USHORT, 1),

		("SmartErrorLog", USHORT, 1),
		("SmartSelfTest", USHORT, 1),
		("MediaSerialNumber", USHORT, 1),
		("MediaCardPassThrough", USHORT, 1),
		("StreamingFeature", USHORT, 1),
		("GpLogging", USHORT, 1),
		("WriteFua", USHORT, 1),
		("WriteQueuedFua", USHORT, 1),
		("WWN64Bit", USHORT, 1),
		("URGReadStream", USHORT, 1),
		("URGWriteStream", USHORT, 1),
		("ReservedForTechReport", USHORT, 2),
		("IdleWithUnloadFeature", USHORT, 1),
		("Reserved4", USHORT, 2),
	]

class NormalSecurityEraseUnit(LittleEndianStructure):
	__slots__ = ()
	_pack_ = 1
	_fields_ = [
		("TimeRequired", USHORT, 15),
		("ExtendedTimeReported", USHORT, 1),
	]

class EnhancedSecurityEraseUnit(LittleEndianStructure):
	__slots__ = ()
	_pack_ = 1
	_fields_ = [
		("TimeRequired", USHORT, 15),
		("ExtendedTimeReported", USHORT, 1),
	]

class PhysicalLogicalSectorSize(LittleEndianStructure):
	__slots__ = ()
	_pack_ = 1
	_fields_ = [
		("LogicalSectorsPerPhysicalSector", USHORT, 4),
		("Reserved0", USHORT, 8),
		("LogicalSectorLongerThan256Words", USHORT, 1),
		("MultipleLogicalSectorsPerPhysicalSector", USHORT, 1),
		("Reserved1", USHORT, 2),
	]

class CommandSetSupportExt(LittleEndianStructure):
	__slots__ = ()
	_pack_ = 1
	_fields_ = [
		("ReservedForDrqTechnicalReport", USHORT, 1),
		("WriteReadVerify", USHORT, 1),
		("WriteUncorrectableExt", USHORT, 1),
		("ReadWriteLogDmaExt", USHORT, 1),
		("DownloadMicrocodeMode3", USHORT, 1),
		("FreefallControl", USHORT, 1),
		("SenseDataReporting", USHORT, 1),
		("ExtendedPowerConditions", USHORT, 1),
		("Reserved0", USHORT, 6),
		("WordValid", USHORT, 2), # shall be 01b
	]

class CommandSetActiveExt(LittleEndianStructure):
	__slots__ = ()
	_pack_ = 1
	_fields_ = [
		("ReservedForDrqTechnicalReport", USHORT, 1),
		("WriteReadVerify", USHORT, 1),
		("WriteUncorrectableExt", USHORT, 1),
		("ReadWriteLogDmaExt", USHORT, 1),
		("DownloadMicrocodeMode3", USHORT, 1),
		("FreefallControl", USHORT, 1),
		("SenseDataReporting", USHORT, 1),
		("ExtendedPowerConditions", USHORT, 1),
		("Reserved0", USHORT, 6),
		("Reserved1", USHORT, 2),
	]

class SecurityStatus(LittleEndianStructure):
	__slots__ = ()
	_pack_ = 1
	_fields_ = [
		("SecuritySupported", USHORT, 1),
		("SecurityEnabled", USHORT, 1),
		("SecurityLocked", USHORT, 1),
		("SecurityFrozen", USHORT, 1),
		("SecurityCountExpired", USHORT, 1),
		("EnhancedSecurityEraseSupported", USHORT, 1),
		("Reserved0", USHORT, 2),
		("SecurityLevel", USHORT, 1),
		("Reserved1", USHORT, 7),
	]

class CfaPowerMode1(LittleEndianStructure):
	__slots__ = ()
	_pack_ = 1
	_fields_ = [
		("MaximumCurrentInMA", USHORT, 12),
		("CfaPowerMode1Disabled", USHORT, 1),
		("CfaPowerMode1Required", USHORT, 1),
		("Reserved0", USHORT, 1),
		("Word160Supported", USHORT, 1),
	]

class DataSetManagementFeature(LittleEndianStructure):
	__slots__ = ()
	_pack_ = 1
	_fields_ = [
		("SupportsTrim", USHORT, 1),
		("Reserved0", USHORT, 15),
	]

class SCTCommandTransport(LittleEndianStructure):
	__slots__ = ()
	_pack_ = 1
	_fields_ = [
		("Supported", USHORT, 1),
		("Reserved0", USHORT, 1),
		("WriteSameSuported", USHORT, 1),
		("ErrorRecoveryControlSupported", USHORT, 1),
		("FeatureControlSuported", USHORT, 1),
		("DataTablesSuported", USHORT, 1),
		("Reserved1", USHORT, 6),
		("VendorSpecific", USHORT, 4),
	]

class BlockAlignment(LittleEndianStructure):
	__slots__ = ()
	_pack_ = 1
	_fields_ = [
		("AlignmentOfLogicalWithinPhysical", USHORT, 14),
		("Word209Supported", USHORT, 1), # shall be set to 1
		("Reserved0", USHORT, 1), # shall be cleared to 0
	]

class NVCacheCapabilities(LittleEndianStructure):
	__slots__ = ()
	_pack_ = 1
	_fields_ = [
		("NVCachePowerModeEnabled", USHORT, 1),
		("Reserved0", USHORT, 3),
		("NVCacheFeatureSetEnabled", USHORT, 1),
		("Reserved1", USHORT, 3),
		("NVCachePowerModeVersion", USHORT, 4),
		("NVCacheFeatureSetVersion", USHORT, 4),
	]

class NVCacheOptions(LittleEndianStructure):
	__slots__ = ()
	_pack_ = 1
	_fields_ = [
		("NVCacheEstimatedTimeToSpinUpInSeconds", UCHAR),
		("Reserved", UCHAR),
	]

class TransportMajorVersion(LittleEndianStructure):
	__slots__ = ()
	_pack_ = 1
	_fields_ = [
		("MajorVersion", USHORT, 12),
		("TransportType", USHORT, 4),
	]

class IDENTIFY_DEVICE_DATA(LittleEndianStructure):
	__slots__ = ()
	_pack_ = 1
	_fields_ = [
		("GeneralConfiguration", GeneralConfiguration), # word 0
		("NumCylinders", USHORT),
		("SpecificConfiguration", USHORT),
		("NumHeads", USHORT),
		("Retired1", USHORT*2),
		("NumSectorsPerTrack", USHORT),
		("VendorUnique1", USHORT*3),
		("SerialNumber", UCHAR*20),
		("Retired2", USHORT*2),
		("Obsolete1", USHORT),
		("FirmwareRevision", UCHAR*8),
		("ModelNumber", UCHAR*40),
		("MaximumBlockTransfer", UCHAR),
		("VendorUnique2", UCHAR),
		("TrustedComputing", TrustedComputing), # word 48
		("Capabilities", Capabilities),
		("ObsoleteWords51", USHORT*2),
		("TranslationFieldsValid", USHORT, 3),
		("Reserved3", USHORT, 5),
		("FreeFallControlSensitivity", USHORT, 8),
		("NumberOfCurrentCylinders", USHORT),
		("NumberOfCurrentHeads", USHORT),
		("CurrentSectorsPerTrack", USHORT), # word 56
		("CurrentSectorCapacity", ULONG),
		("CurrentMultiSectorSetting", UCHAR),
		("MultiSectorSettingValid", UCHAR, 1),
		("ReservedByte59", UCHAR, 3),
		("SanitizeFeatureSupported", UCHAR, 1),
		("CryptoScrambleExtCommandSupported", UCHAR, 1),
		("OverwriteExtCommandSupported", UCHAR, 1),
		("BlockEraseExtCommandSupported", UCHAR, 1),
		("UserAddressableSectors", ULONG),
		("ObsoleteWord62", USHORT),
		("MultiWordDMASupport", USHORT, 8),
		("MultiWordDMAActive", USHORT, 8),
		("AdvancedPIOModes", USHORT, 8),
		("ReservedByte64", USHORT, 8),
		("MinimumMWXferCycleTime", USHORT),
		("RecommendedMWXferCycleTime", USHORT),
		("MinimumPIOCycleTime", USHORT),
		("MinimumPIOCycleTimeIORDY", USHORT),
		("AdditionalSupported", AdditionalSupported),
		("ReservedWords70", USHORT*5), # word 74
		("QueueDepth", USHORT, 5),
		("ReservedWord75", USHORT, 11),
		("SerialAtaCapabilities", SerialAtaCapabilities),
		("SerialAtaFeaturesSupported", SerialAtaFeaturesSupported),
		("SerialAtaFeaturesEnabled", SerialAtaFeaturesEnabled),
		("MajorRevision", USHORT),
		("MinorRevision", USHORT),
		("CommandSetSupport", CommandSetSupport),
		("CommandSetActive", CommandSetActive),
		("UltraDMASupport", USHORT, 8),
		("UltraDMAActive", USHORT, 8),
		("NormalSecurityEraseUnit", NormalSecurityEraseUnit),
		("EnhancedSecurityEraseUnit", EnhancedSecurityEraseUnit), # word 90
		("CurrentAPMLevel", USHORT, 8),
		("ReservedWord91", USHORT, 8),
		("MasterPasswordID", USHORT),
		("HardwareResetResult", USHORT),
		("CurrentAcousticValue", USHORT, 8),
		("RecommendedAcousticValue", USHORT, 8),
		("StreamMinRequestSize", USHORT),
		("StreamingTransferTimeDMA", USHORT),
		("StreamingAccessLatencyDMAPIO", USHORT),
		("StreamingPerfGranularity", ULONG),
		("Max48BitLBA", ULONG*2),
		("StreamingTransferTime", USHORT),
		("DsmCap", USHORT),
		("PhysicalLogicalSectorSize", PhysicalLogicalSectorSize), # 106
		("InterSeekDelay", USHORT),
		("WorldWideName", USHORT*4),
		("ReservedForWorldWideName128", USHORT*4),
		("ReservedForTlcTechnicalReport", USHORT),
		("WordsPerLogicalSector", USHORT*2),
		("CommandSetSupportExt", CommandSetSupportExt),
		("CommandSetActiveExt", CommandSetActiveExt),
		("ReservedForExpandedSupportandActive", USHORT*6), # 126
		("MsnSupport", USHORT, 2),
		("ReservedWord127", USHORT, 14),
		("SecurityStatus", SecurityStatus),
		("ReservedWord129", USHORT*31),
		("CfaPowerMode1", CfaPowerMode1), # word 160
		("ReservedForCfaWord161", USHORT*7),
		("NominalFormFactor", USHORT, 4),
		("ReservedWord168", USHORT, 12),
		("DataSetManagementFeature", DataSetManagementFeature),
		("AdditionalProductID", USHORT*4),
		("ReservedForCfaWord174", USHORT*2),
		("CurrentMediaSerialNumber", USHORT*30),
		("SCTCommandTransport", SCTCommandTransport), # 206
		("ReservedWord207", USHORT*2),
		("BlockAlignment", BlockAlignment),
		("WriteReadVerifySectorCountMode3Only", USHORT*2),
		("WriteReadVerifySectorCountMode2Only", USHORT*2),
		("NVCacheCapabilities", NVCacheCapabilities),
		("NVCacheSizeLSW", USHORT),
		("NVCacheSizeMSW", USHORT),
		("NominalMediaRotationRate", USHORT),
		("ReservedWord218", USHORT),
		("NVCacheOptions", NVCacheOptions),
		("WriteReadVerifySectorCountMode", USHORT, 8),
		("ReservedWord220", USHORT, 8),
		("ReservedWord221", USHORT),
		("TransportMajorVersion", TransportMajorVersion), # 222
		("TransportMinorVersion", USHORT),
		("ReservedWord224", USHORT*6),
		("ExtendedNumberOfUserAddressableSectors", ULONG*2),
		("MinBlocksPerDownloadMicrocodeMode03", USHORT),
		("MaxBlocksPerDownloadMicrocodeMode03", USHORT),
		("ReservedWord236", USHORT*19),
		("Signature", USHORT, 8), # word 255
		("CheckSum", USHORT, 8),
	]

# SMART sub command list
IDE_SMART_READ_ATTRIBUTES = 0xD0
IDE_SMART_READ_THRESHOLDS = 0xD1
IDE_SMART_ENABLE_DISABLE_AUTOSAVE = 0xD2
IDE_SMART_SAVE_ATTRIBUTE_VALUES = 0xD3
IDE_SMART_EXECUTE_OFFLINE_DIAGS = 0xD4
IDE_SMART_READ_LOG = 0xD5
IDE_SMART_WRITE_LOG = 0xD6
IDE_SMART_ENABLE = 0xD8
IDE_SMART_DISABLE = 0xD9
IDE_SMART_RETURN_STATUS = 0xDA
IDE_SMART_ENABLE_DISABLE_AUTO_OFFLINE = 0xDB

from ctypes import sizeof
assert sizeof(IDENTIFY_DEVICE_DATA()) == 512



from __future__ import absolute_import, division, print_function, unicode_literals

from builtins import map

from itertools import chain

import numpy as np
from sklearn.preprocessing import LabelEncoder
from nltk.tokenize import word_tokenize

from .func import identity

class BatchLabelEncoder(object):

	""" Similar to `sklearn.preprocessing.LabelEncoder` but accepts a list of sentences as inputs
		and returns a list of lists of integer labels.
		Warning: Calling `fit` doesn't reset the encoder. To do so call `reset()` explicitly.
		cf. `keras_preprocessing.text.Tokenizer`
	"""

	def __init__(self, tokenizer):
		# type: (str, ) -> None

		self.tokenizer = {
			"none": identity,
			"nltk": word_tokenize,
		}[tokenizer]
		self.reset()

	def reset(self):
		# type: () -> None

		""" Initializes all mappings with empty containers. """

		self.token2idx = {} # type: Dict[Any, int]
		self.idx2count = []
		self.idx2token = []

	def finalize(self, vocab_size):
		# type: (int, ) -> None

		""" This only keeps the most frequent `vocab_size` words in the vocabulary.
			This will change the assigned indices, so if `transform` was used before `finalize`,
			they cannot be transformed back using `inverse_transform`.
		"""

		self.idx2count = np.array(self.idx2count)
		self.idx2token = np.array(self.idx2token)
		indices = np.argsort(-self.idx2count)[:vocab_size]

		self.idx2count = self.idx2count[indices]
		self.idx2token = self.idx2token[indices]

		self.token2idx = {token: i for i, token in enumerate(self.idx2token)}

	def partial_fit_single(self, token):
		# type: (T, ) -> int

		vocab_size = self.num_labels
		idx = self.token2idx.setdefault(token, vocab_size)
		if idx == vocab_size: # inserted new token
			self.idx2count.append(1)
			self.idx2token.append(token)
		else:
			self.idx2count[idx] += 1

		return idx

	def partial_fit(self, sentence):
		# type: (str, ) -> None

		for token in self.tokenizer(sentence):
			self.partial_fit_single(token)

	def partial_fit_batch(self, sentences):
		# type: (Iterable[str], ) -> None

		for sentence in sentences:
			self.partial_fit(sentence)

	def fit_single(self, token):
		# type: (str, ) -> None

		self.partial_fit_single(token)

	def fit(self, sentence):
		# type: (str, ) -> None

		self.partial_fit(sentence)

	def fit_batch(self, sentences):
		# type: (Iterable[str], ) -> None

		self.partial_fit_batch(sentences)

	def transform_single(self, token):
		# type: (Iterable[str], ) -> Iterator[List[int]]

		""" Unknown labels are skipped. """

		return self.token2idx[token]

	def transform(self, sentence, ignore=True):
		# type: (str, ) -> Iterator[int]

		for token in self.tokenizer(sentence):
			try:
				yield self.transform_single(token)
			except KeyError:
				if ignore:
					pass
				else:
					raise

	def transform_batch(self, sentences, ignore=True):
		# type: (Iterable[str], ) -> Iterator[List[int]]

		""" Unknown labels are skipped. """

		for sentence in sentences:
			yield list(self.transform(sentence, ignore))

	def fit_transform_single(self, token):
		# type: (Iterable[str], ) -> Iterator[List[int]]

		return self.partial_fit_single(token)

	def fit_transform(self, sentences):
		# type: (str, ) -> Iterator[int]

		self.partial_fit(sentences)
		return self.transform(sentences)

	def fit_transform_batch(self, sentences):
		# type: (Iterable[str], ) -> Iterator[List[int]]

		self.partial_fit_batch(sentences)
		return self.transform_batch(sentences)

	def inverse_transform_single(self, idx):
		# type: (int, ) -> str

		return self.idx2token[idx]

	def inverse_transform(self, indices):
		# type: (Iterable[int], ) -> Iterator[str]

		for idx in indices:
			yield self.inverse_transform_single(idx)

	def inverse_transform_batch(self, list_of_indices):
		# type: (Iterable[Iterable[int]], ) -> Iterator[List[str]]

		for indices in list_of_indices:
			yield list(self.inverse_transform(indices))

	@property
	def num_labels(self):
		assert len(self.token2idx) == len(self.idx2count) == len(self.idx2token)
		return len(self.idx2token)



from __future__ import absolute_import, division, print_function, unicode_literals

from typing import TYPE_CHECKING

from .func import identity

if TYPE_CHECKING:
	from typing import Any, Callable

def opjit(*args, **kwargs):
	# type: (*Any, **Any) -> Callable

	""" Optional / opportunistic numba jit decorator. """

	try:
		from numba import njit

	except ImportError:
		from warnings import warn
		warn("Numba not found. Using slower pure Python version.", stacklevel=2)

		return identity

	kwargs.setdefault("fastmath", True)
	kwargs.setdefault("cache", True)

	def dec(func):
		return njit(*args, **kwargs)(func)

	return dec



from __future__ import absolute_import, division, print_function, unicode_literals

import logging

from math import sqrt

# was: FindWidthToHeightRatio
def perspective_rectangle_aspect_ratio(corners, principal_point, focal_length=None):

	""" Return the width-to-height ratio of a perspectivly transformed rectangle.
		taken from: http://stackoverflow.com/questions/1194352/proportions-of-a-perspective-deformed-rectangle
		corners: corners of the quadrilateral
		principal point: center of the image (not the quadrilateral)
	"""

	sqr = lambda x: x**2
	#(m1x, m1y), (m2x, m2y), (m3x, m3y), (m4x, m4y) = corners
	(m1x, m1y), (m3x, m3y), (m4x, m4y), (m2x, m2y) = corners
	u0, v0 = principal_point

	# in case it matters: licensed under GPLv2 or later
	# legend:
	# sqr(x)  = x*x
	# sqrt(x) = square root of x

	# let m1x,m1y ... m4x,m4y be the (x,y) pixel coordinates
	# of the 4 corners of the detected quadrangle
	# i.e. (m1x, m1y) are the cordinates of the first corner,
	# (m2x, m2y) of the second corner and so on.
	# let u0, v0 be the pixel coordinates of the principal point of the image
	# for a normal camera this will be the center of the image,
	# i.e. u0=IMAGEWIDTH/2; v0 =IMAGEHEIGHT/2
	# This assumption does not hold if the image has been cropped asymmetrically

	# first, transform the image so the principal point is at (0,0)
	# this makes the following equations much easier
	m1x = m1x - u0
	m1y = m1y - v0
	m2x = m2x - u0
	m2y = m2y - v0
	m3x = m3x - u0
	m3y = m3y - v0
	m4x = m4x - u0
	m4y = m4y - v0

	# temporary variables k2, k3
	k2 = ((m1y - m4y)*m3x - (m1x - m4x)*m3y + m1x*m4y - m1y*m4x) / ((m2y - m4y)*m3x - (m2x - m4x)*m3y + m2x*m4y - m2y*m4x)
	k3 = ((m1y - m4y)*m2x - (m1x - m4x)*m2y + m1x*m4y - m1y*m4x) / ((m3y - m4y)*m2x - (m3x - m4x)*m2y + m3x*m4y - m3y*m4x)

	# f_squared is the focal length of the camera, squared
	# if k2==1 OR k3==1 then this equation is not solvable
	# if the focal length is known, then this equation is not needed
	# in that case assign f_squared= sqr(focal_length)
	if focal_length:
		f_squared = sqr(focal_length)
	else:
		f_squared = -((k3*m3y - m1y)*(k2*m2y - m1y) + (k3*m3x - m1x)*(k2*m2x - m1x)) / ((k3 - 1)*(k2 - 1))

	try:
		logging.info("Focal length: {}".format(sqrt(f_squared)))
	except ValueError:
		logging.warning("estimating FL failed, {}, k2: {}, k3: {}, f^2: {}".format([(m1x, m1y), (m2x, m2y), (m3x, m3y), (m4x, m4y)], k2, k3, f_squared))

	try:
		# The width/height ratio of the original rectangle
		whRatio = sqrt(
			(sqr(k2 - 1) + sqr(k2*m2y - m1y)/f_squared + sqr(k2*m2x - m1x)/f_squared) /
			(sqr(k3 - 1) + sqr(k3*m3y - m1y)/f_squared + sqr(k3*m3x - m1x)/f_squared)
		)
	except ValueError:
		logging.warning("estimating AR failed, {}, k2: {}, k3: {}, f^2: {}".format([(m1x, m1y), (m2x, m2y), (m3x, m3y), (m4x, m4y)], k2, k3, f_squared))
		return -1

	# if k2==1 AND k3==1, then the focal length equation is not solvable
	# but the focal length is not needed to calculate the ratio.
	# I am still trying to figure out under which circumstances k2 and k3 become 1
	# but it seems to be when the rectangle is not distorted by perspective,
	# i.e. viewed straight on. Then the equation is obvious:
	if (k2==1 and k3==1):
		whRatio = sqrt(
			(sqr(m2y-m1y) + sqr(m2x-m1x)) /
			(sqr(m3y-m1y) + sqr(m3x-m1x))
		)

	# After testing, I found that the above equations
	# actually give the height/width ratio of the rectangle,
	# not the width/height ratio.
	# If someone can find the error that caused this,
	# I would be most grateful.
	# until then:
	return whRatio
	whRatio = 1/whRatio




from itertools import chain
from typing import TYPE_CHECKING

from networkx import Graph, connected_components

if TYPE_CHECKING:
	import numpy as np

def complete_weighted_bipartite_graph(first, second, weights):
	# type: (Sequence[T], Sequence[T], np.ndarray) -> Graph

	graph = Graph()

	graph.add_nodes_from(first, bipartite=0)
	graph.add_nodes_from(second, bipartite=1)

	for i in range(len(first)):
		for j in range(len(second)):
			graph.add_edge(first[i], second[j], weight=weights[i][j])

	return graph

def connected_subgraph(graph, minsize=2):
	nodelist = list(chain.from_iterable(comp for comp in connected_components(graph) if len(comp) >= minsize))
	return graph.subgraph(nodelist)



from __future__ import absolute_import, division, print_function, unicode_literals

from functools import partial
from datetime import datetime

from msgpack import ExtType, packb, unpackb

from .compat.datetime import datetime as compat_datetime

class EXTID:
	datetime = 1
	set = 2
	frozenset = 3

def default(obj):
	if isinstance(obj, (datetime, compat_datetime)):
		return ExtType(EXTID.datetime, obj.isoformat().encode("utf-8")) # todo: use struct.Struct instead
	elif isinstance(obj, set):
		return ExtType(EXTID.set, packb(tuple(obj), use_bin_type=True))
	elif isinstance(obj, frozenset):
		return ExtType(EXTID.frozenset, packb(tuple(obj), use_bin_type=True))

	return obj

def ext_hook(code, data):
	if code == EXTID.datetime:
		return compat_datetime.fromisoformat(data.decode("utf-8"))
	elif code == EXTID.set:
		return set(unpackb(data, use_list=False, raw=False))
	elif code == EXTID.frozenset:
		return frozenset(unpackb(data, use_list=False, raw=False))

	return data

dumps = partial(packb, default=default, use_bin_type=True)
loads = partial(unpackb, ext_hook=ext_hook, use_list=False, raw=False)



from __future__ import absolute_import, division, print_function, unicode_literals

from builtins import map
from future.utils import PY2, viewkeys, viewitems

import re
from itertools import chain
from locale import strxfrm
from functools import partial
from typing import TYPE_CHECKING

from .iter import switched_enumerate

if TYPE_CHECKING:
	from typing import Any, Dict, Iterable, Tuple, IO

english_consonants = "bcdfghjklmnpqrstvwxz" # y?
english_vowels = "aeiouy" # y?

german_consonants = "bcdfghjklmnpqrstvwxyz"
german_vowels = "aeiou"

#def cycleleft(s, start=0, end=-1): # bad
#	return s[:start] + s[start+1:end] + s[start] + s[end+1:]

#def cycleright(s, start=0, end=-1): # bad
#	return s[:start] + s[end] + s[start:end] + s[end+1:]

def tryint(obj):
	try:
		return int(obj)
	except (ValueError, TypeError):
		return obj

def locale_sorted(seq, case_insensitive=True, lower_before_upper=True):
	if case_insensitive:
		if lower_before_upper:
			key = lambda s: (strxfrm(s.lower()), strxfrm(s.swapcase()))
		else:
			key = lambda s: (strxfrm(s.lower()), strxfrm(s))
	else:
		if lower_before_upper:
			key = lambda s: (strxfrm(s.swapcase()), strxfrm(s))
		else:
			key = strxfrm

	return sorted(seq, key=key)

def build_multiple_replace(d, escape=True):
	# type: (Dict[str, str], bool) -> Callable

	if escape:
		it = map(re.escape, viewkeys(d))
	else:
		it = viewkeys(d)

	cp = re.compile("(" + "|".join(it) + ")")
	return partial(cp.sub, lambda m: d[m.group(0)])

def deunicode(s):
	# type: (str, ) -> str

	pairs = (("\u2014", "--"), ("\u2017", ">="), ("\u2019", "'"), ("\u2026", "..."))
	s = replace_pairs(s, pairs)
	s.encode("ascii") #check for further unicode
	return s

def are_parentheses_matched(s, open="([{", close=")]}"):
	# type: (str, str, str) -> bool

	stack = list()
	parentheses = dict(chain(switched_enumerate(open), switched_enumerate(close)))
	assert len(parentheses) == len(open) + len(close)

	for char in s:
		if char in open:
			stack.append(parentheses[char])
		if char in close:
			try:
				if stack.pop() != parentheses[char]: # wrong nesting
					return False
			except IndexError: # stack empty
				return False

	return len(stack) == 0 # unclosed parentheses left

# was: cond_join
def filter_join(s, it, func=None):
	# type: (str, Iterable[str], Optional[Callable]) -> str

	if func is None:
		return s.join(i for i in it if i)
	else:
		return s.join(i for i in it if func(i))

def surrounding_join(j, it, l="", r=""):
	# type: (str, Iterable[str], str, str) -> str

	""" Example: surrounding_join(", ", ("a", "b", "c"), l="[", r="]") -> "[a], [b], [c]"
	"""

	s = (r+j+l).join(it)
	if s != "":
		return l+s+r
	return ""

def replace_pairs(s, items):
	# type: (str, Iterable[Tuple[str, str]]) -> str

	for key, value in items:
		s = s.replace(key, value)
	return s

def replace_pairs_bytes(s, items):
	# type: (bytes, Dict[bytes, Optional[bytes]]) -> bytes

	frm = b"".join(k for k, v in viewitems(items) if v)
	to = b"".join(v for k, v in viewitems(items) if v)
	delete = b"".join(k for k, v in viewitems(items) if v is None)

	if PY2:
		import string
		table = string.maketrans(frm, to)
	else:
		table = s.maketrans(frm, to)

	return s.translate(table, delete)

def replace_pairs_chars(s, items):
	# type: (str, Dict[str, Optional[str]]) -> str

	if PY2:
		# table = s.maketrans(items) # 'unicode' object has no attribute 'maketrans'
		table = {ord(k):v for k, v in viewitems(items)}
	else:
		table = s.maketrans(items)

	return s.translate(table)

def replace_list(s, character_list, replace_char):
	# type: (str, Sequence[str], str) -> str

	""" should work with CaseInsensitiveString also and replace replace_list_insensitive"""

	for find_char in character_list:
		s = s.replace(find_char, replace_char)
	return s



from __future__ import absolute_import, division, print_function, unicode_literals

from builtins import bytes, zip
from itertools import cycle

def xor_stream(it, key, cyclekey=True):
	# type: (Iterable[bytes], Iterable[bytes]) -> Iterator[bytes]

	if cyclekey:
		key = cycle(key)

	for i, j in zip(it, key):
		yield bytes([i ^ j])



from __future__ import absolute_import, division, print_function, unicode_literals

from future.utils import PY2
import platform
from typing import TYPE_CHECKING

from .os_shared import is_os_64bit

if TYPE_CHECKING:
	from typing import Callable
	from builtins import str

system = platform.system()

def rename(func_name):
	# type: (str, ) -> Callable

	if PY2:
		func_name = func_name.encode("ascii")

	def decorator(func):
		func.__name__ = func_name
		return func

	return decorator

def _not_available(func_name):
	# type: (str) -> None

	@rename(func_name)
	def inner(*args, **kwargs):
		raise OSError("{}() is not available for {}".format(func_name, system))

	return inner

if system == "Windows":

	from .os_win import _lock as lock
	from .os_win import _unlock as unlock
	from .os_win import _disk_usage_windows as disk_usage
	from .os_win import _volume_info_windows as volume_info
	from .os_win import _filemanager_cmd_windows as filemanager_cmd
	from .os_win import _get_appdata_dir as get_appdata_dir
	from .os_win import _islink as islink
	from .os_win import _uncabspath as uncabspath
	from .os_win import _interrupt_windows as interrupt

elif system == "Linux":

	from .os_posix import _lock as lock
	from .os_posix import _unlock as unlock
	from .os_posix import _disk_usage_posix as disk_usage
	volume_info = _not_available("volume_info")
	from .os_posix import _filemanager_cmd_posix as filemanager_cmd
	get_appdata_dir = _not_available("get_appdata_dir")
	from os.path import islink
	from os.path import abspath as uncabspath
	from .os_posix import _interrupt_posix as interrupt

elif system == "Darwin":

	from .os_posix import _lock as lock
	from .os_posix import _unlock as unlock
	from .os_posix import _disk_usage_posix as disk_usage
	volume_info = _not_available("volume_info")
	from .os_mac import _filemanager_cmd_mac as filemanager_cmd
	get_appdata_dir = _not_available("get_appdata_dir")
	from os.path import islink
	from os.path import abspath as uncabspath
	from .os_posix import _interrupt_posix as interrupt

else:
	lock = _not_available("lock")
	unlock = _not_available("unlock")
	disk_usage = _not_available("disk_usage")
	volume_info = _not_available("volume_info")
	filemanager_cmd = _not_available("filemanager_cmd")
	get_appdata_dir = _not_available("get_appdata_dir")
	islink = _not_available("islink")
	uncabspath = _not_available("uncabspath")
	interrupt = _not_available("interrupt")

lock.__doc__ = """ Locks access to the file (on Posix) or its contents (Windows). """
unlock.__doc__ = """ Unlocks access to the file. """
disk_usage.__doc__ = """ Returns (total, used, free) bytes on disk. """
volume_info.__doc__ = """ filesystem and name of the volume """
filemanager_cmd.__doc__ = """ Returns a shell command that when executed starts the file manager of the OS. """
get_appdata_dir.__doc__ = """ Returns the roaming appdata directory of the current user. """
islink.__doc__ = """ islink """
uncabspath.__doc__ = """ uncabspath """
interrupt.__doc__ = """ interrupt """

import sys
if sys.version_info >= (3, 3):
	from shutil import disk_usage



from __future__ import absolute_import, division, print_function, unicode_literals

import json
from itertools import islice
from typing import TYPE_CHECKING

from .file import copen

if TYPE_CHECKING:
	from typing import Any, Callable, Dict, Optional, TextIO, Union

if __debug__:
	import jsonschema

class BuiltinEncoder(json.JSONEncoder):
	def default(self, obj):
		from datetime import date

		if isinstance(obj, (set, frozenset)):
			return tuple(obj)
		elif isinstance(obj, complex):
			return [obj.real, obj.imag]
		elif isinstance(obj, date):
			return obj.isoformat()

		return json.JSONEncoder.default(self, obj)

def read_json_schema(path):
	# type: (str, ) -> Dict[str, Any]

	with open(path, "r", encoding="utf-8") as fr:
		return json.load(fr)

def read_json(path, schema=None, object_hook=None):
	# type: (str, Optional[Union[str, dict]]) -> Any

	""" Read the json file at `path` and optionally validates the input according to `schema`.
		The validation requires `jsonschema`.
		`schema` can either be a path as well, or a Python dict which represents the schema.
		`object_hook` is passed through to `json.load`.
	"""

	with copen(path, "rt", encoding="utf-8") as fr:
		obj = json.load(fr, object_hook=object_hook)

	if schema is None:
		return obj

	from jsonschema import validate

	if isinstance(schema, str):
		schema = read_json_schema(schema)

	validate(obj, schema)
	return obj

class json_lines(object):

	""" Read and write files in the JSON Lines format (http://jsonlines.org).
	"""

	def __init__(self,
		stream, doclose,
		cls=None, object_hook=None, parse_float=None, parse_int=None, parse_constant=None, object_pairs_hook=None, **kw
	):
		# type: (TextIO, bool, Optional[json.JSONEncoder], Optional[Callable], Optional[Callable], Optional[Callable], Optional[Callable], Optional[Callable], **Any) -> None

		""" Don't use directly. Use `from_path` or `from_stream` classmethods instead. """

		""" fixme: how should `close` be handled?
			1: If you don't want to close `stream`, just don't call `close()` or use as context manager.
			2: use `doclose` argument to decide
		"""

		self.f = stream
		self.doclose = doclose
		self.newline = "\n"

		self.json_kwargs = {
			"cls": cls,
			"object_hook": object_hook,
			"parse_float": parse_float,
			"parse_int": parse_int,
			"parse_constant": parse_constant,
			"object_pairs_hook": object_pairs_hook,
		}
		self.json_kwargs.update(kw)

	@staticmethod
	def from_path(
		file, mode="r", encoding="UTF-8", errors="strict", newline=None,
		cls=None, object_hook=None, parse_float=None, parse_int=None, parse_constant=None, object_pairs_hook=None, **kw
	):
		# type: (str, str, int, str, str, Optional[str], bool, Optional[Callable], Optional[json.JSONEncoder], Optional[Callable], Optional[Callable], Optional[Callable], Optional[Callable], Optional[Callable], **Any) -> json_lines

		stream = copen(file, mode, encoding=encoding, errors=errors, newline=newline)
		return json_lines(stream, True, cls, object_hook, parse_float, parse_int, parse_constant, object_pairs_hook, **kw)

	@staticmethod
	def from_stream(
		stream,
		cls=None, object_hook=None, parse_float=None, parse_int=None, parse_constant=None, object_pairs_hook=None, **kw
	):
		# type: (TextIO, Optional[json.JSONEncoder], Optional[Callable], Optional[Callable], Optional[Callable], Optional[Callable], Optional[Callable], **Any) -> json_lines

		return json_lines(stream, False, cls, object_hook, parse_float, parse_int, parse_constant, object_pairs_hook, **kw)

	def __enter__(self):
		# type: () -> None

		return self

	def __exit__(self, exc_type, exc_value, traceback):
		self.close()

	def iterrange(self, start=0, stop=None):
		# type: (int, Optional[int]) -> Iterator

		linenum = start+1
		try:
			for line in islice(self.f, start, stop):
				line = line.rstrip().lstrip("\x00") # fixme: strip \0 is only a temp fix!
				if line:
					yield json.loads(line, **self.json_kwargs)
				linenum += 1

		except ValueError as e: # json.JSONDecodeError in python 3.5+
			e.lineno = linenum
			logger.error("JSON Lines parse error in line %s: '%r'", linenum, line)
			raise

	def __iter__(self):
		# type: () -> Iterator

		return self.iterrange()

	def write(self, obj, skipkeys=False, ensure_ascii=False, check_circular=True, allow_nan=True,
		cls=None, indent=None, separators=None, default=None, sort_keys=False, **kw):
		# type: (Any, bool, bool, bool, bool, Optional[Callable], Optional[str], Optional[Tuple[str, str]], Optional[Callable], bool **Any) -> None
		line = json.dumps(obj, skipkeys=skipkeys, ensure_ascii=ensure_ascii, check_circular=check_circular,
			allow_nan=allow_nan, cls=cls, indent=indent, separators=separators, default=default, sort_keys=sort_keys,
			**kw) + self.newline
		self.f.write(line)

	def close(self):
		# type: () -> None

		if self.doclose:
			self.f.close()



from __future__ import absolute_import, division, print_function, unicode_literals

from copy import copy

def cast(object, class_, instanceof=object, *args, **kwargs):
	""" Changes the class of `object` to `class_` if `object` is an instance of `instanceof`,
		calls the initializer and returns it.
	"""

	object = copy(object)
	if isinstance(object, instanceof):
		object.__class__ = class_
		object.__init__(*args, **kwargs)
	else:
		raise TypeError("Object is not an instance of {}".format(instanceof.__name__))
	return object



from __future__ import absolute_import, division, print_function, unicode_literals

import os, signal, logging
from ctypes import byref, sizeof, create_unicode_buffer, WinError, cast, c_wchar_p, memset
from ctypes.wintypes import ULARGE_INTEGER, LPCWSTR, DWORD
from msvcrt import get_osfhandle
from typing import TYPE_CHECKING

from cwinsdk.shared.ehstorioctl import MAX_PATH
from cwinsdk.shared.ntdef import PWSTR
from cwinsdk.um.consoleapi import SetConsoleMode, GetConsoleMode, ENABLE_VIRTUAL_TERMINAL_PROCESSING
from cwinsdk.um.consoleapi2 import CONSOLE_SCREEN_BUFFER_INFO, GetConsoleScreenBufferInfo
from cwinsdk.um.fileapi import INVALID_FILE_ATTRIBUTES, GetDiskFreeSpaceExW, GetVolumeInformationW, LockFileEx, UnlockFileEx, GetFileAttributesW
from cwinsdk.um.processenv import GetStdHandle
from cwinsdk.um.combaseapi import CoTaskMemFree
from cwinsdk.um.ShlObj_core import SHGetKnownFolderPath
from cwinsdk.um.KnownFolders import FOLDERID_RoamingAppData
from cwinsdk.um.minwinbase import LOCKFILE_EXCLUSIVE_LOCK, LOCKFILE_FAIL_IMMEDIATELY, OVERLAPPED
from cwinsdk.um.WinBase import STD_OUTPUT_HANDLE
from cwinsdk.um.winnt import FILE_ATTRIBUTE_REPARSE_POINT

from .os_shared import _usagetuple, _volumeinfotuple

if TYPE_CHECKING:
	from typing import Tuple
	from ctypes.wintypes import HANDLE

unc_prefix = "\\\\?\\"

def get_stdout_handle():
	# type: () -> HANDLE

	""" Might return a redirect handle. For the real handle, use CreateFile("CONOUT$") """

	return GetStdHandle(STD_OUTPUT_HANDLE)

class EnableAnsi(object): # doesn't work for some reason...

	def __init__(self):
		# type: () -> None

		#os.system('') # calls cmd, which sets ANSI mode, but doesn't disable it when exiting (it's a bug probably)
		self.handle = get_stdout_handle()
		self.oldmode = DWORD()

		GetConsoleMode(self.handle, byref(self.oldmode))
		SetConsoleMode(self.handle, self.oldmode.value|ENABLE_VIRTUAL_TERMINAL_PROCESSING)

	def close(self):
		SetConsoleMode(self.handle, self.oldmode.value)

	def __enter__(self):
		return None

	def __exit__(self, *args):
		self.close()

def _islink(path):
	""" Tests if `path` refers to a symlink or a junction.
		- Python >= 3.2 `os.path.islink()` only supports symlinks, not junctions.
		- Python < 3.2 `os.path.islink()` always returns `False` on Windows.
		This function works in all cases.
	"""

	FileName = os.fspath(path)
	FileAttributes = GetFileAttributesW(FileName)

	if FileAttributes == INVALID_FILE_ATTRIBUTES:
		raise WinError()
	return FileAttributes & FILE_ATTRIBUTE_REPARSE_POINT == FILE_ATTRIBUTE_REPARSE_POINT

def _uncabspath(path):
	# type: (str, )-> str

	return unc_prefix+os.path.abspath(path)

def _lock(fp, exclusive=True, block=False):
	# type: (IO, bool, bool) -> None

	fd = fp.fileno()
	handle = get_osfhandle(fd)

	flags = 0
	if exclusive:
		flags |= LOCKFILE_EXCLUSIVE_LOCK
	if not block:
		flags |= LOCKFILE_FAIL_IMMEDIATELY

	overlapped = OVERLAPPED()
	memset(byref(overlapped), 0, sizeof(overlapped))

	LockFileEx(handle, flags, 0, 0xffffffff, 0xffffffff, overlapped)

def _unlock(fp):
	# type: (IO, bool, bool) -> None

	fd = fp.fileno()
	handle = get_osfhandle(fd)

	overlapped = OVERLAPPED()
	memset(byref(overlapped), 0, sizeof(overlapped))

	UnlockFileEx(handle, 0, 0xffffffff, 0xffffffff, overlapped)

def _get_appdata_dir():
	KF_FLAG_CREATE = 0x00008000

	rfid = byref(FOLDERID_RoamingAppData)
	Flags = 0
	Token = None
	Path = PWSTR()

	try:
		result = SHGetKnownFolderPath(rfid, KF_FLAG_CREATE, None, byref(Path))
	except OSError:
		logging.error("SHGetKnownFolderPath result: %X".format(result & 0xffffffff))
		raise

	ret = cast(Path, c_wchar_p).value
	CoTaskMemFree(Path)
	return ret

def _disk_usage_windows(path):
	# type: (str, ) -> _usagetuple

	DirectoryName = LPCWSTR(path)
	FreeBytesAvailableToCaller = ULARGE_INTEGER(0) # user free
	TotalNumberOfBytes = ULARGE_INTEGER(0) # user total
	TotalNumberOfFreeBytes = None # total free

	GetDiskFreeSpaceExW(DirectoryName, byref(FreeBytesAvailableToCaller), byref(TotalNumberOfBytes), TotalNumberOfFreeBytes)
	return _usagetuple(
		TotalNumberOfBytes.value,
		TotalNumberOfBytes.value-FreeBytesAvailableToCaller.value,
		FreeBytesAvailableToCaller.value
	)

def _volume_info_windows(path):
	# type: (str, ) -> _volumeinfotuple

	assert path.endswith("\\"), "X: usually doesn't work. X:\\ does."

	RootPathName = LPCWSTR(path)
	VolumeNameBuffer = create_unicode_buffer(MAX_PATH+1)
	VolumeNameSize = MAX_PATH+1
	VolumeSerialNumber = DWORD()
	MaximumComponentLength = DWORD()
	FileSystemFlags = DWORD()
	FileSystemNameBuffer = create_unicode_buffer(MAX_PATH+1)
	FileSystemNameSize = MAX_PATH+1

	GetVolumeInformationW(
		RootPathName,
		VolumeNameBuffer,
		VolumeNameSize,
		byref(VolumeSerialNumber),
		byref(MaximumComponentLength),
		byref(FileSystemFlags),
		FileSystemNameBuffer,
		FileSystemNameSize
	)

	return _volumeinfotuple(
		VolumeNameBuffer.value,
		VolumeSerialNumber.value,
		MaximumComponentLength.value,
		FileSystemFlags.value,
		FileSystemNameBuffer.value
	)

def _interrupt_windows():
	os.kill(os.getpid(), signal.CTRL_C_EVENT)

def _filemanager_cmd_windows(path):
	# type: (str, ) -> str

	return 'explorer.exe /select,"{}"'.format(path)

if __name__ == "__main__":
	s = '\033[35m'+'color-test'+'\033[39m'+" test end"
	print(s)
	with EnableAnsi():
		print(s)
	print(s)



from __future__ import absolute_import, division, print_function, unicode_literals
from builtins import range

from random import choice, randrange

def randstr(length, charset):
	return "".join(choice(charset) for i in range(length))

def randbytes(size):
	return b"".join(chr(randrange(0, 256)) for _ in range(size))

def rgb_colors():
	while True:
		rgb = randrange(0, 256**3)
		rg, b = divmod(rgb, 256)
		r, g = divmod(rg, 256)
		yield (r, g, b)



from __future__ import absolute_import, division, print_function, unicode_literals

from future.moves.urllib.parse import quote, unquote

from base64 import urlsafe_b64decode, urlsafe_b64encode
from typing import TYPE_CHECKING

from flask import request
from werkzeug.routing import BaseConverter

if TYPE_CHECKING:
	from typing import Optional

""" examples

user_dict = {
	"username": "password"
}

@before_request
def auth():
	return do_basic_auth(user_dict)

"""

class Base64Converter(BaseConverter):

	def to_python(self, value): # decode once as the wsgi app already receives a decoded url
		# type: (str, ) -> str

		return urlsafe_b64decode(value.encode("ascii")).decode("utf-8")

	def to_url(self, value):
		# type: (str, ) -> str

		return urlsafe_b64encode(value.encode("utf-8")).decode("ascii")

class DoubleQuoteConverter(BaseConverter):

	def to_python(self, value): # decode once as the wsgi app already receives a decoded url
		# type: (str, ) -> str

		return unquote(value)

	def to_url(self, value):
		# type: (str, ) -> str

		return quote(quote(value, safe=""), safe="")

def check_auth(username, password, user_dict):
	# type: (H, T, Dict[H, T]) -> bool

	try:
		return user_dict[username] == password
	except KeyError:
		return False

def authenticate():
	# type: () -> Response

	"""Sends a 401 response that enables basic auth"""

	return Response(
		'Could not verify your access level for that URL.\n'
		'You have to login with proper credentials', 401,
		{'WWW-Authenticate': 'Basic realm="Login Required"'}
	)

def do_basic_auth(user_dict):
	# type: (dict, ) -> Optional[bool]

	auth = request.authorization
	if not auth or not check_auth(auth.username, auth.password, user_dict):
		return authenticate()
	else:
		return None



from __future__ import absolute_import, division, print_function, unicode_literals

from typing import TYPE_CHECKING

from OpenSSL import crypto # pyOpenSSL
from Crypto.PublicKey import RSA # pycryptodome

if TYPE_CHECKING:
	from OpenSSL.crypto import X509
	from Crypto.PublicKey.RSA import RsaKey

def load_certificate_file(filename):
	# type: (str, ) -> X509

	with open(filename, "rb") as fr:
		return crypto.load_certificate(crypto.FILETYPE_PEM, fr.read())

def get_pubkey_from_x509(x509):
	# type: (X509, ) -> bytes

	return crypto.dump_privatekey(crypto.FILETYPE_ASN1, x509.get_pubkey())

def load_rsa_keyfile(filename):
	# type: (str, ) -> RsaKey

	with open(filename, "rb") as fr:
		return RSA.importKey(fr.read())

def generate_rsa_keyfile_pair(priv_key, pub_key, bits=1024*4):
	# type: (str, str, int) -> RsaKey

	key = RSA.generate(bits)

	with open(priv_key, "wb") as fw:
		fw.write(key.exportKey())

	with open(pub_key, "wb") as fw:
		fw.write(key.publickey().exportKey())

	return key



from __future__ import absolute_import, division, print_function, unicode_literals

from math import sqrt

from .numpy import issquare
from .numba import opjit

@opjit(["void(f4[:,:], f4[:], f4)", "void(f8[:,:], f8[:], f8)"])
def _choldate(L, x, sign):

	for k in range(L.shape[0]):
		r = sqrt(L[k, k]*L[k, k] + sign * x[k]*x[k])
		c = r / L[k, k]
		s = x[k] / L[k, k]
		L[k, k] = r

		L[k+1:, k] = (L[k+1:, k] + sign * s * x[k+1:]) / c
		x[k+1:] = c * x[k+1:] - s * L[k+1:, k]

def choldate(L, x, sign):
	# type: (np.ndarray, np.ndarray, str) -> None

	""" This function computes the lower triangular Cholesky decomposition L' of matrix A' from L in-place
		(the cholesky decomp of A) where A' = A + sign*x*x^T.
		Based on: https://en.wikipedia.org/wiki/Cholesky_decomposition#Rank-one_update

		L: square lower triangular matrix
	"""

	if not issquare(L) or x.shape != (L.shape[0], ):
		raise ValueError("Invalid dimensions")

	try:
		sign = {"+": +1., "-": -1.}[sign]
	except KeyError:
		raise ValueError("Invalid sign")

	return _choldate(L, x, sign)

if __name__ == "__main__":
	import timeit
	from .numpy import random_triangular_matrix

	size = 100
	L = random_triangular_matrix(size)
	x = np.random.uniform(0, 1, size)

	choldate(L, x, "+") # warmup
	print(min(timeit.repeat('choldate(L, x, "+")', number=10000, repeat=5, globals=globals())))



from __future__ import absolute_import, division, print_function, unicode_literals

class SplitRecursive(object):

	"""Fast non-prime-based factorial algorithm.
		Python port of: http://www.luschny.de/math/factorial/java/FactorialSplit.java.html
	"""

	def factorial(self, n):
		# type: (int, ) -> int

		if n < 0:
			raise ArithmeticError("Factorial: n has to be >= 0, but was " + n)

		if n < 2:
			return 1

		p = 1
		r = 1
		self.N = 1

		log2n = n.bit_length() - 1 # should be the same as: int(floor(log(n, 2)))
		h = 0
		shift = 0
		high = 1

		while h != n:
			shift += h
			h = n >> log2n
			log2n -= 1
			length = high
			if (h & 1) == 1:
				high = h
			else:
				high = h - 1
			length = (high - length) >> 1

			if length > 0:
				p *= self.product(length)
				r *= p
		return r << shift

	def product(self, n):
		m = n >> 1
		if m == 0:
			self.N += 2
			return self.N
		if n == 2:
			self.N += 2
			ret = self.N * (self.N + 2)
			self.N += 2
			return ret
		return self.product(n - m) * self.product(m)



from __future__ import absolute_import, division, print_function, unicode_literals

from future.utils import viewitems
from collections import defaultdict
from typing import TYPE_CHECKING

if TYPE_CHECKING:
	from typing import Any, Callable, Dict, Hashable, Iterable, Mapping, Optional

	T = TypeVar("T")
	U = TypeVar("U")
	H = TypeVar("H", bound=Hashable)

def hasvalues(d):
	return {k:v for k, v in viewitems(d) if v}

def itemgetter(it):
	# type: (Iterable[T], ) -> Callable[[Mapping[T, U]], Iterator[U]]

	""" Similar to `operator.itemgetter` except that it always expects and returns iterables.
		Compare `mapmap`
	"""

	return lambda d: (d[i] for i in it)

def subdict(d, it):
	# type: (Mapping[T, U], Iterable[T]) -> Dict[T, U]

	""" Uses the elements of `it` as keys to extract a new sub-dictionary. """

	return {key: d[key] for key in it}

def subdictdefault(d, it, default=None):
	# type: (Mapping[T, U], Iterable[T], Optional[T]) -> Dict[T, U]

	""" Uses the elements of `it` as keys to extract a new sub-dictionary. """

	return {key: d.get(key, default) for key in it}

class keydefaultdict(defaultdict):

	""" defaultdict which passes a key to the default factory. """

	def __missing__(self, key):
		# type: (Hashable, ) -> Any

		if self.default_factory is None:
			raise KeyError(key)
		else:
			value = self[key] = self.default_factory(key)
			return value



from __future__ import absolute_import, division, print_function, unicode_literals

from future.utils import viewkeys
import json

from genutility.http import URLRequest

class StreamWatcher(object):

	def __init__(self, api):
		self.api = api
		self.followed = {name: False for name in api.get_followed()}

	def watch(self, notify_started, notify_stopped):
		for name in viewkeys(self.followed):
			stream = self.api.get_stream(name)

			if stream is not None and not self.followed[name]:
				self.followed[name] = True
				notify_started(name)
			elif stream is None and self.followed[name]:
				self.followed[name] = False
				notify_stopped(name)

class TwitchAPI(object):

	base = "https://api.twitch.tv/kraken/"
	follows = "users/{}/follows/channels?client_id={}"
	streams = "streams/{}?client_id={}"

	def __init__(self, username, client_id):
		self.username = username
		self.client_id = client_id

	def get_followed(self):
		url = TwitchAPI.base + TwitchAPI.follows.format(self.username, self.client_id)
		data = URLRequest(url).load()
		d = json.loads(data)
		return [follow["channel"]["_links"]["self"].rsplit("/", 1)[1] for follow in d["follows"]]

	def get_stream(self, channel):
		url = TwitchAPI.base + TwitchAPI.streams.format(channel, self.client_id)
		data = URLRequest(url).load()
		d = json.loads(data)
		return d["stream"]

	def watcher(self):
		return StreamWatcher(self)



from __future__ import absolute_import, division, print_function, unicode_literals

from .exceptions import NoResult, InconsistentState

def findone(db, query):
	docs = list(db.find(query))

	if len(docs) == 0:
		raise NoResult("No result found")
	elif len(docs) == 1:
		return docs[0]
	else:
		raise InconsistentState("More than one result found")



from __future__ import absolute_import, division, print_function, unicode_literals

from builtins import str

from io import open
from typing import TYPE_CHECKING

import toml

if TYPE_CHECKING:
	from typing import Any

def read_toml(path):
	# type: (str, ) -> Any

	with open(path, "r", encoding="utf-8") as fr:
		return toml.load(fr)

def write_toml(obj, path):
	# type: (str, Any) -> None

	with open(path, "w", encoding="utf-8") as fw:
		toml.dump(obj, fw)



from __future__ import absolute_import, division, print_function, unicode_literals

import bs4

def simple_microdata_parser(bstree): # schema.org Microdata, replace with `extruct` module?
	# type: (bs4.BeautifulSoup, ) -> dict

	def is_item_scope(elem):
		return elem.has_attr("itemscope") and elem.has_attr("itemtype") and not elem.has_attr("itemprop")

	def is_item_scopeprop(elem):
		return elem.has_attr("itemscope") and elem.has_attr("itemtype") and elem.has_attr("itemprop")

	def is_item_prop(elem):
		return elem.has_attr("itemprop") and not elem.has_attr("itemscope") and not elem.has_attr("itemtype")

	def rec(tree, dict=None):
		if dict is None:
			dict = {}

		for child in tree.children:
			if isinstance(child, bs4.element.Tag):
				if is_item_scope(child):
					key = child["itemtype"]
					value = rec(child)
					#print("tag {} is scope: {}".format(child.name, child["itemtype"]))

					try:
						dict[key].append(value)
					except KeyError:
						dict[key] = [value]

				elif is_item_scopeprop(child):
					#print("tag {} is scopeprop: {}, {}".format(child.name, child["itemtype"], child["itemprop"]))

					try:
						dict[(child["itemtype"], child["itemprop"])].append(rec(child))
					except KeyError:
						dict[(child["itemtype"], child["itemprop"])] = [rec(child)]

				elif is_item_prop(child):
					key = child["itemprop"]

					if child.name == "meta":
						value = child["content"]
					elif child.contents:
						value = child.contents[0].string
					else:
						continue

					try:
						dict[key].append(value)
					except KeyError:
						dict[key] = [value]

				else:
					rec(child, dict)

		return dict

	return rec(bstree) #.html



import csv, fnmatch
from itertools import islice
from io import open

import re2 as re

from .time import PrintStatementTime

# UNFINISHED!!! use re2 or hyperscan for pathological regex

class Browscap(object):

	def __init__(self, path="browscap.csv"):
		self.path = path

	@staticmethod
	def convert_to_regex(pattern):
		ret = "^{}$".format(re.escape(pattern)) # fnmatch.translate(pattern)
		return ret.replace("\\?", ".").replace("\\*", ".*")

	def iter_patterns(self):
		with open(self.path, "rt", newline="") as fr:
			csvreader = csv.reader(fr)
			for row in islice(csvreader, 2, None):
				yield self.convert_to_regex(row[0]), row[6]

	def match_v1(self, useragent):
		data = list((re.compile(p, re.IGNORECASE), t) for p, t in self.iter_patterns())

		with PrintStatementTime():
			for pattern, type in data:
				if pattern.match(useragent):
					print(type)

	def match_v2(self, useragent):
		pattern = "|".join(p for p, t in self.iter_patterns())
		pattern = re.compile(pattern, re.IGNORECASE, max_mem=1024**3)

		with PrintStatementTime():
			if pattern.match(useragent):
				print(True)

if __name__ == "__main__":
	from argparse import ArgumentParser
	
	parser = ArgumentParser()
	parser.add_argument("path", help="browscap.csv file")
	args = parser.parse_args()

	cap = Browscap(args.path)
	#cap.match_v1("Firefox")
	cap.match_v2("Firefox")



from __future__ import absolute_import, division, print_function, unicode_literals

class ParseError(Exception):
	pass

class AccessDenied(Exception):
	pass

class TemporaryError(Exception):
	pass

class IteratorExhausted(Exception):
	pass

# results, control flow

class Skip(Exception):
	pass

class NoResult(Exception):
	pass

# external errors

class InconsistentState(Exception):
	pass

# values, input errors

class EmptyIterable(ValueError):
	pass

class MalformedFile(ValueError):
	pass



from __future__ import absolute_import, division, print_function, unicode_literals

from enum import Enum

from genutility.func import identity

class BisectRetVal(Enum):
	Lower = -1
	Equal = 0
	Higher = 1

def make_binary_search(target):
	# type: (int, ) -> Callable[[int], BisectRetVal]

	def binary_search_func(x):
		if x < target:
			return BisectRetVal.Higher
		elif x > target:
			return BisectRetVal.Lower
		else:
			return BisectRetVal.Equal
	return binary_search_func

def make_binary_search_sequence(seq, target, key=None):
	# type: (Sequence, Any) -> Callable[[int], BisectRetVal]

	key = key or identity

	def binary_search_func(x):
		result = key(seq[x])
		if result < target:
			return BisectRetVal.Higher
		elif result > target:
			return BisectRetVal.Lower
		else:
			return BisectRetVal.Equal
	return binary_search_func

def bisect_left_generic(lo, hi, func):
	# type: (int, int, Callable[[int], BisectRetVal]) -> int

	""" Generic left bisection (binary search). Finds a specific `lo<=x<=hi` where
		`func(x) == BisectRetVal.Equal` or where the return value changes from BisectRetVal.Higher
		to `BisectRetVal.Lower`.
	"""

	while lo < hi:
		mid = (lo + hi) // 2
		result = func(mid)
		if result == BisectRetVal.Higher:
			lo = mid+1
		else:
			hi = mid
	return lo

def bisect_right_generic(lo, hi, func):
	# type: (int, int, Callable[[int], BisectRetVal]) -> int

	""" Generic right bisection (binary search). Finds a specific `lo<=x<=hi` where
		`func(x) == BisectRetVal.Equal` or where the return value changes from BisectRetVal.Higher
		to `BisectRetVal.Lower`.
	"""

	while lo < hi:
		mid = (lo + hi) // 2
		result = func(mid)
		if result == BisectRetVal.Lower:
			hi = mid
		else:
			lo = mid+1
	return lo

def bisect_left_sequence(seq, target, key=None):
	""" see: bisect.bisect_left """

	func = make_binary_search_sequence(seq, target, key)
	return bisect_left_generic(0, len(seq), func)

def bisect_right_sequence(seq, target, key=None):
	""" see: bisect.bisect_right """

	func = make_binary_search_sequence(seq, target, key)
	return bisect_right_generic(0, len(seq), func)

def search_sorted(seq, newitem, key=None):
	# type: (Sequence[T], U, Callable[[T], U]) -> int

	key = key or identity

	for i, item in enumerate(seq):
		if newitem < key(item):
			return i

	return len(seq)

if __name__ == "__main__":
	seq = [1, 2, 3, 4, 4, 5]
	from bisect import bisect_left, bisect_right
	for target in (1, 2, 3, 4, 5, 0.5, 1.5, 2.5, 3.5, 4.5, 5.5):
		print(bisect_left_sequence(seq, target) == bisect_left(seq, target))
		print(search_sorted(seq, target) == bisect_right_sequence(seq, target) == bisect_right(seq, target))



from __future__ import absolute_import, division, print_function, unicode_literals

import threading, logging
from queue import Queue, Empty
from typing import TYPE_CHECKING

from .exceptions import NoResult

if TYPE_CHECKING:
	from typing import Any, Callable, Iterable, Optional, Tuple, Union

logger = logging.getLogger(__name__)

class Worker(threading.Thread):

	TASK_IDLE = 0
	TASK_COMPLETE = 1
	TASK_EXCEPTION = 2

	def __init__(self, tasks, returns):
		# type: (Queue[Optional[Tuple[int, int, Callable, tuple, dict]]], Queue[Tuple[int, int, Tuple[int, Union[Any, Exception]]]]) -> None

		threading.Thread.__init__(self)
		self.tasks = tasks
		self.returns = returns

		self.daemon = True
		self.running = True
		self.start()

	def run(self):
		# type: () -> None

		while self.running:
			#self.returns.put((self.TASK_IDLE, self))
			task = self.tasks.get(True)
			if task:
				state, id, func, args, kargs = task
			else:
				self.tasks.task_done()
				self.running = False
				break

			try:
				ret = func(*args, **kargs)
			except Exception as e: # pylint: disable=broad-except
				logger.exception("threaded task failed")
				self.returns.put((state, self.TASK_EXCEPTION, (id, e)))
			else:
				self.returns.put((state, self.TASK_COMPLETE, (id, ret)))

			try:
				self.tasks.task_done()
			except ValueError: # if the task was canceled inbetween
				pass

		logger.debug("thread ended")

	def cancel(self):
		# type: () -> None

		self.running = False

class ThreadPool(object):

	def __init__(self, num_threads):
		# type: (int, ) -> None

		self.num_threads = num_threads
		self.tasks = Queue() # type: Queue[Optional[Tuple[int, int, Callable, tuple, dict]]]
		self.returns = Queue() # type: Queue[Tuple[int, int, Tuple[int, Union[Any, Exception]]]]
		self.my_worker = [Worker(self.tasks, self.returns) for _ in range(num_threads)]
		self.state = 0 # this is used to be able to cancel (or ignore the results of ongoing) tasks

	def add_task(self, id, func, *args, **kargs):
		# type: (int, Callable[..., Any], *Any, **Any) -> None

		self.tasks.put((self.state, id, func, args, kargs))

	def get(self):
		# type: () -> Tuple[int, Optional[Any]]

		while True: # so invalid states can be ignored
			state, type, result = self.returns.get(True) # Tuple[int, int, Tuple[int, Union[Any, Exception]]]
			if state == self.state:
				if type == Worker.TASK_COMPLETE:
					return result # (id, ret)
				else:
					id, e = result
					return id, None # maybe raise here?
			# else: old result, ignore

	@staticmethod
	def clear(queue):
		# type: (Queue[Any], ) -> None

		""" Only undocumented methods. """

		with queue.mutex:
			queue.queue.clear()
			queue.all_tasks_done.notify_all()
			queue.unfinished_tasks = 0

	@staticmethod
	def _clear(queue):
		# type: (Queue[Any], ) -> None

		""" Only uses documented methods, but is slower. """

		while not queue.empty():
			queue.get()
			queue.task_done()

	def close(self, finishall=False):
		# type: (bool, ) -> None

		""" if `finishall` is True, it will finish all tasks currently in the queue,
			if `finishall` is False, only currently active tasks will be finished
		"""
		if finishall:
			for i in range(self.num_threads):
				self.tasks.put(None)
		else:
			for w in self.my_worker:
				w.cancel()

	def cancel(self):
		# type: () -> None

		""" Does not cancel already running tasks. Just clears queue and ignores
			results of currently running tasks. """

		self.clear(self.tasks)
		self.clear(self.returns)
		self.state += 1

	def wait(self):
		self.tasks.join()

def gather_all_unsorted(threadpool, func, params, *args, **kwargs):
	# type: (ThreadPool, Callable, Iterable, tuple, dict) -> Iterator[Tuple[Any, Any]]

	""" Runs multiple tasks concurrently and returns all results in the order
		of execution finish as soon as possible.
	"""

	for i, param in enumerate(params):
		threadpool.add_task(i, func, param, *args, **kwargs)

	for i in range(len(params)):
		yield threadpool.get()

def gather_any(threadpool, func, params, *args, **kwargs):
	# type: (ThreadPool, Callable, Iterable, tuple, dict) -> Tuple[Any, Any]

	""" Runs multiple tasks concurrently and just returns the result of the first
		task which finishes.
	"""

	for id, result in gather_all_unsorted(threadpool, func, params, *args, **kwargs):
		threadpool.cancel()
		return id, result # result will be None, if tasked threw exception
	raise NoResult("No task returned a valid result")

def NotThreadSafe(verify=False):
	# type: (bool, ) -> Callable[type]

	""" Class decorator. Only works on new style classes (object)
		`verify=True` prohibits write access to attributes.
	"""

	def OnClass(TheClass):
		# type: (type, ) -> type

		if verify:
			orig_init = TheClass.__init__
			orig_setattr = TheClass.__setattr__

			def __setattr__(self, name, value):
				if self.thread_ident == threading.current_thread().ident:
					orig_setattr(self, name, value)
				else:
					raise RuntimeError("{}.{} = {} called from different thread".format(TheClass.__name__, name, value))

			def __init__(self, *args, **kws):
				orig_setattr(self, "thread_ident", threading.current_thread().ident)
				orig_init(self, *args, **kws)

			TheClass.__init__ = __init__
			TheClass.__setattr__ = __setattr__
		return TheClass
	return OnClass

class AbortIteration(Exception):
	pass

class IterWorker(threading.Thread):

	STATE_WAITING = 0
	STATE_RUNNING = 1
	STATE_PAUSING = 2
	STATE_PAUSED = 3
	STATE_STOPPED = 4

	CMD_RESUME = 0
	CMD_ABORT = 1

	def __init__(self, taskqueue, onstatechange=None):
		# type: (Queue, Optional[Callable]) -> None

		threading.Thread.__init__(self)
		self.queue = taskqueue
		self.onstatechange = onstatechange
		self.control = Queue()
		self.state = self.STATE_WAITING

	@property
	def state(self):
		# type: () -> int

		return self._state

	@state.setter
	def state(self, value):
		# type: (int, ) -> None

		assert value in [0, 1, 2, 3, 4]
		self._state = value
		if self.onstatechange:
			self.onstatechange(value)

	def run(self):
		# type: () -> None

		while True:
			iter = self.queue.get()
			if not iter:
				self.state = self.STATE_STOPPED
				return
			self.state = self.STATE_RUNNING
			try:
				for __ in iter:
					if self.state == self.STATE_PAUSING:
						self.state = self.STATE_PAUSED
						cmd = self.control.get()
						if cmd == self.CMD_RESUME:
							self.state = self.STATE_RUNNING
							continue
						elif cmd == self.CMD_ABORT:
							break
			except AbortIteration as e:
				logger.warning("Task aborted")
			except Exception as e:
				logger.exception("Task threw exception")
			self.queue.task_done() #untested
			self.state = self.STATE_WAITING

	def pause(self):
		# type: () -> None

		if self.state == self.STATE_RUNNING:
			self.state = self.STATE_PAUSING

	def resume(self):
		# type: () -> None

		if self.state == self.STATE_PAUSED:
			self.control.put(self.CMD_RESUME)

	def stop(self):
		# type: () -> None

		if self.state < self.STATE_STOPPED:
			self.queue.put(None)

	def clear(self):
		# type: () -> None

		try:
			while True:
				self.queue.get_nowait()
				self.queue.task_done()
		except Empty:
			pass

	def abort(self):
		# type: () -> None

		self.pause()
		self.control.put(self.CMD_ABORT)

#was: DownloadWorker
class ProgressWorker(threading.Thread):
	def __init__(self, manager, tasks):
		threading.Thread.__init__(self)
		self.manager = manager
		self.tasks = tasks

		self.running = True
		self.task, self.done, self.total = None, None, None

	def run(self):
		while self.running:
			self.task = self.tasks.get()
			if self.task:
				func, args, kwargs = self.task
			else:
				break

			try:
				result = func(self.report, *args, **kwargs)
			except Exception:
				logger.exception("Task failed")
				result = None
			self.task, self.done, self.total = None, None, None
			self.manager.finalize(result)

	def quit(self):
		self.running = False

	def report(self, done, total):
		self.done = done
		self.total = total

	def get(self):  # not using locks atm, can return inconsistent data
		if self.task and self.done and self.total:
			return (self.task, self.done, self.total)
		else:
			return None

#was: ThreadedDownloader
class ProgressThreadPool(object):

	def __init__(self, concurrent=1):
		# from config
		self.concurrent = concurrent

		self.waiting_queue = Queue()
		self.completed = []
		self.failed = []
		self.lock = threading.Lock()

		self.workers = list(ProgressWorker(self, self.waiting_queue) for i in range(self.concurrent))
		for w in self.workers:
			w.daemon = True # program will end even if threads are still running
			w.start()

	def start(self, callable, *args, **kwargs):
		self.waiting_queue.put((callable, args, kwargs))

	def finalize(self, result):
		if result:
			setter, status, localname, length = result
			with self.lock:
				if status:
					self.failed.append((status, localname, length))
				else:
					self.completed.append((localname, length))
					setter(localname)
		else: # task failed
			pass

	def clear_completed(self):
		self.completed = [] # assignment is atomic right?

	def cancel_pending(self):
		# copied from https://stackoverflow.com/a/18873213
		with self.waiting_queue.mutex:
			self.waiting_queue.queue.clear()
			self.waiting_queue.all_tasks_done.notify_all()
			self.waiting_queue.unfinished_tasks = 0

	def get_waiting(self):
		return list(task for task in self.waiting_queue.queue if task)

	def get_completed(self):
		return self.completed

	def get_failed(self):
		return self.failed

	def get_running(self): # no locks, inconsistent data
		return list(task for task in (w.get() for w in self.workers) if task)

class ThreadsafeList(list): # untested!!!

	""" This is a list object with context manager to handle locking to perform multiple operations
		on a list in a threadsafe way.
		Example.
		l = ThreadsafeList([1, 2, 3])
		l.append(4) # atomic, no locking needed
		with l:
			l[0] += 4 # not atomic, thus executed with context manager
	"""

	def __init__(self, it):
		list.__init__(self, it)
		self.lock = threading.Lock()

	def __enter__(self):
		self.lock.acquire()
		return self

	def __exit__(self, exc_type, exc_value, traceback):
		self.lock.release()



from __future__ import absolute_import, division, print_function, unicode_literals

import logging, asyncio
from typing import TYPE_CHECKING

import aiohttp
from orderedset import OrderedSet

from .datetime import now

if TYPE_CHECKING:
	from typing import Any

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

class DownloadTask(object):

	def __init__(self, url, path="."):
		self.url = url
		self.path = path
		self.downloaded = 0
		self.resumable = False
		self.dt_started = None
		self.dt_finished = now()

	def __hash__(self):
		return hash((self.url, self.path))

	def start(self):
		logger.info("starting download")
		self.dt_started = now()

	def done(self):
		logger.info("finished download")
		self.dt_finished = now()

class DownloadManager(object):

	def __init__(self):
		self.loop = asyncio.get_event_loop()
		self.timeout = aiohttp.ClientTimeout(total=60)
		self.session = aiohttp.ClientSession(loop=self.loop, timeout=self.timeout, auto_decompress=False)
		self.concurrent_downloads = 3
		#self.sem = asyncio.Semaphore(1000)
		self.chunksize = 1024*1024 # file write buffer

		self.queue = OrderedSet()
		self.active = OrderedSet()
		self.done = OrderedSet()

	def status(self):
		return "Queued: {}, active: {}, done: {}".format(len(self.queue), len(self.active), len(self.done))

	def _enqueue(self, task, priority):
		self.queue.add(task)

	def _start(self, task):
		self.active.add(task)
		task = asyncio.ensure_future(self._download(task))
		return task

	def _trystart(self):
		if len(self.active) < self.concurrent_downloads:
			try:
				task = self.queue.pop()
				self._start(task)
			except KeyError:
				if not self.active:
					logger.info("all done")
					#self.loop.stop()
					#task = asyncio.ensure_future(self._close())

	async def _download(self, task):
		task.start()
		#await asyncio.sleep(10)

		# send http head request first to check for range support
		
		#async with self.session.get(task.url, headers={"Range": "bytes=0-10"}) as response:
		async with self.session.get(task.url, headers={}) as response:
			stream = response.content
			try:
				size = int(response.headers.get('content-length'))
			except (ValueError, TypeError):
				size = None

			accept_range = response.headers.get('Accept-Ranges', 'none').lower()
			print(accept_range)

			if response.status == 200: # range not supported
				pass
			elif response.status == 206: # range supported
				assert accept_range == "bytes"
				bytes_range = response.headers.get('Content-Range') # 'bytes 0-10/46239'

			with open(task.path, "wb", buffering=self.chunksize) as fd:
				async for data in stream.iter_any():
					task.downloaded += len(data)
					fd.write(data)

			if size and size != task.downloaded:
				print("incomplete", task.downloaded, "of", size)

		task.done()

		self.active.remove(task)
		self.done.add(task)
		self._trystart()

	def download(self, url, path="tmp.txt", priority=0, force=False):
		logger.info("starting download")
		task = DownloadTask(url, path)
		if force:
			self._start(task)
		else:
			self._enqueue(task, priority)
			self._trystart()

	async def _close(self):
		await self.session.close()

if __name__ == "__main__":

	dm = DownloadManager()

	import wx
	from wxasync import AsyncBind, WxAsyncApp, StartCoroutine
	from asyncio.events import get_event_loop
	import time

	class TestFrame(wx.Frame):
		def __init__(self, parent=None):
			super(TestFrame, self).__init__(parent)
			vbox = wx.BoxSizer(wx.VERTICAL)
			button1 =  wx.Button(self, label="Submit")
			self.edit =  wx.StaticText(self, style=wx.ALIGN_CENTRE_HORIZONTAL|wx.ST_NO_AUTORESIZE)
			self.edit_timer =  wx.StaticText(self, style=wx.ALIGN_CENTRE_HORIZONTAL|wx.ST_NO_AUTORESIZE)
			vbox.Add(button1, 2, wx.EXPAND|wx.ALL)
			vbox.AddStretchSpacer(1)
			vbox.Add(self.edit, 1, wx.EXPAND|wx.ALL)
			vbox.Add(self.edit_timer, 1, wx.EXPAND|wx.ALL)
			self.SetSizer(vbox)
			self.Layout()
			AsyncBind(wx.EVT_BUTTON, self.async_callback, button1)
			StartCoroutine(self.update_clock, self)

		async def async_callback(self, event):
			self.edit.SetLabel("WX WAITING")
			dm.download("http://google.com") # doesn't support range
			#dm.download("https://curl.haxx.se/docs/manpage.html") # supports range
			self.edit.SetLabel("WX COMPLETE")

		async def update_clock(self):
			while True:
				self.edit_timer.SetLabel(dm.status())
				await asyncio.sleep(0.5)

	app = WxAsyncApp()
	frame = TestFrame()
	frame.Show()
	app.SetTopWindow(frame)
	loop = get_event_loop()
	loop.run_until_complete(app.MainLoop())



from __future__ import absolute_import, division, print_function, unicode_literals

from builtins import filter, range

import sys, os, os.path, subprocess, logging

from ..twothree.filesystem import tofs, fromfs
from ..string import surrounding_join

logger = logging.getLogger(__name__)

class RarError(Exception):
	def __init__(self, msg, returncode=None, cmd=None, output=""):
		Exception.__init__(self, msg)
		self.returncode = returncode
		self.cmd = cmd
		self.output = output

"""WinRAR Exit Codes
255 	USER BREAK	User stopped the process
9	CREATE ERROR	Create file error
8	MEMORY ERROR	Not enough memory for operation
7	USER ERROR	Command line option error
6	OPEN ERROR	Open file error
5	WRITE ERROR	Write to disk error
4	LOCKED ARCHIVE	Attempt to modify an archive previously locked by the 'k' command
3	CRC ERROR	A CRC error occurred when unpacking
2	FATAL ERROR	A fatal error occurred
1	WARNING	Non fatal error(s) occurred
0	SUCCESS	Successful operation (User exit)"""

class Rar(object):

	windows_executable = "C:/Program Files/WinRAR/Rar.exe"

	def __init__(self, archive, executable=None):
		"""archive: archive to work with, everything which is supported by winrar
		executable: path/to/Rar.exe"""

		self.exe = executable or self.windows_executable

		if not os.path.isfile(self.exe):
			raise ValueError("Invalid executable")

		self.archive = archive
		self.filelist = []
		self.cmd = ""
		self.flags = {}
		self.flags["lock"] = [False, "k"]
		self.flags["delete"] = [False, "df"]
		self.flags["test"] = [False, "t"]
		self.flags["filetime"] = [False, "tl"]
		self.flags["append_archive_name"] = [False, "ad"]
		self.options = {}
		self.options["split"] = [False, "v%s"]
		self.options["compression"] = [False, "m%u"]
		self.options["password"] = [False, "p%s"]
		self.options["password_header"] = [False, "hp%s"]
		self.options["recovery"] = [False, "rr%up"]
		self.options["recovery_volumes"] = [False, "rr%u%%"]

	def add_file(self, pathname):
		self.filelist.append(pathname)

	def add_files(self, filelist):
		self.filelist.extend(filelist)

	def set_compression(self, level):
		"""level: 0 store, 1 fastest, 2 fast, 3 normal, 4 good, 5 best (default: 3)"""

		if level not in range(0, 6) and level is not False:
			raise RarError("Invalid parameter: Set compression level (0-store...3-default...5-best)")
		self.options["compression"][0] = level

	def set_password(self, password, encrypt_filenames=False):
		if encrypt_filenames:
			self.options["password_header"][0] = password
			self.options["password"][0] = False
		else:
			self.options["password"][0] = password
			self.options["password_header"][0] = False

	def add_recovery_info(self, rr):
		"""rr: recovery record in percent, only 1-10 is valid"""

		if rr < 1 or rr > 10:
			raise RarError("Only 1%-10% valid")
		self.options["recovery"][0] = rr

	def add_recovery_volumes(self, rv):
		"""rv: number of recovery volumes"""

		self.options["recovery_volumes"][0] = rv

	def split(self, split):
		self.options["split"][0] = split

	def lock(self, flag = True):
		self.flags["lock"][0] = flag

	def delete_after_archiving(self, flag = True):
		"""flag (bool): delete files after archiving"""

		self.flags["delete"][0] = flag

	def test_after_archiving(self, flag = True):
		"""flag (bool): test archive after archiving"""

		self.flags["test"][0] = flag

	def set_archive_to_filetime(self, flag = True):
		self.flags["filetime"][0] = flag

	def commandline(self, cmd):
		self.cmd = cmd.strip()

	def execute(self, args):
		cmd = "{} {}".format(self.exe, args)
		logger.debug("CMD: " + cmd)
		try:
			ret = subprocess.check_output(tofs(cmd), stderr=subprocess.STDOUT, cwd=os.getcwd())
		except UnicodeEncodeError as e:
			raise RarError("UnicodeError, Win32Console fault")
		except subprocess.CalledProcessError as e:
			raise RarError("Error", e.returncode, fromfs(e.cmd), e.output.decode(sys.stdout.encoding)) #should use only stderr

	def test(self, password="-"):
		self.execute('t -p{} "{}"'.format(password, self.archive))

	def get_files_str(self):
		return surrounding_join(" ", self.filelist, "\"", "\"")

	def get_flag_str(self):
		return surrounding_join(" ", [value[1] for value in self.flags.itervalues() if value[0] is True], "-", "")

	def get_options_str(self):
		return surrounding_join(" ", [value[1] % value[0] for value in self.options.itervalues() if value[0] is not False], "-", "")

	def get_args(self, command):
		return "%s %s %s %s \"%s\" %s" % (command, self.get_flag_str(), self.get_options_str(), self.cmd, self.archive, self.get_files_str())

	def create(self):
		self.execute(self.get_args("a"))

	def extract(self, dir, mode = 0):
		self.filelist = [dir]
		if mode == 1:
			self.flags["append_archive_name"][0] = True
		self.execute(self.get_args("x"))

	def close(self):
		pass

def create_rar_from_folder(path, dest_path=".", profile_setter_func=None, filter_func=lambda x:True, name_transform=lambda x:x):
	if not os.path.isdir(path):
		return False

	root, dirname = os.path.split(path)

	if dest_path == ".":
		dest_path = root

	cwd = os.getcwd()
	os.chdir(path)

	try:
		r = Rar(os.path.join(dest_path, "{}.rar".format(name_transform(dirname))))
		if profile_setter_func:
			profile_setter_func(r)
		for c in filter(filter_func, os.listdir(".")): #was: listdir_rec
			r.add_file(c)
		r.create()
	except RarError as e:
		logger.error("%s\n%s" % (str(e), e.output))
		return False

	os.chdir(cwd)
	return True

def create_rar_from_file(path, dest_path=".", profile_setter_func=None, name_transform = lambda x:x):
	root, dirname = os.path.split(path)

	if dest_path == ".":
		dest_path = root

	cwd = os.getcwd()
	os.chdir(root)

	try:
		r = Rar(os.path.join(dest_path, "{}.rar".format(name_transform(dirname))))
		if profile_setter_func:
			profile_setter_func(r)
		r.add_file(dirname)
		r.create()
	except RarError as e:
		logger.error("%s\n%s" % (str(e), e.output))
		return False

	os.chdir(cwd)
	return True



from __future__ import absolute_import, division, print_function, unicode_literals

import struct, zlib

png_sig =  b"\x89PNG\r\n\x1a\n"

def png_chunk(type, data=""):
	# http://www.w3.org/TR/PNG/#5Chunk-layout
	length = struct.pack("!I", len(data))
	crc = zlib.crc32(type)
	crc = zlib.crc32(data, crc)
	crc = struct.pack("!i", crc)
	return length + type + data + crc

def IHDR(width, height):
	bitdepth = 1
	colortype = 0 #grayscale
	compression = 0
	filter = 0
	interlace = 0
	data = struct.pack("!IIBBBBB", width, height, bitdepth, colortype, compression, filter, interlace)

	return png_chunk("IHDR", data)

def IEND():
	return png_chunk("IEND", "")

def IDAT(binary, level=9):
	#ignores filter for scanlines
	return png_chunk("IDAT", zlib.compress(binary, level))

def binary2png(binary, width, height):
	return png_sig+IHDR(width, height)+IDAT(binary)+IEND()

__all__ = ["binary2png"]



from __future__ import absolute_import, division, print_function, unicode_literals

from itertools import islice
from typing import TYPE_CHECKING

import numpy as np
from gensim.models.keyedvectors import Vocab
from gensim.models.keyedvectors import KeyedVectors as KeyedVectorsOriginal

from .file import PathOrTextIO

if TYPE_CHECKING:
	from typing import Iterator, Tuple, Optional, Callable, Collection, TextIO

class DuplicateEntry(ValueError):
	pass

class KeyedVectors(KeyedVectorsOriginal):

	""" Enhancement of the original `gensim` `KeyedVectors` class. Supports loading of `glove` and
		`muse` word vector files.
	"""

	@classmethod
	def load_a_format(cls, fin, vocab_size, vector_size, datatype=np.float32, discard=None):
		# type: (TextIO, int, int, Optional[Callable[[str], bool]]) -> KeyedVectorsOriginal

		if discard is None:
			discard = lambda x: False

		result = cls(vector_size)
		result.vector_size = vector_size
		result.vectors = np.zeros((vocab_size, vector_size), dtype=datatype)

		for line in fin:
			word, vect = line.rstrip().split(" ", 1)

			if discard(word):
				continue

			weights = np.fromstring(vect, sep=" ", dtype=np.float32)
			# raise ValueError("invalid vector on line %s; is vector_size incorrect or file otherwise damaged?" % (i+1,))

			if word in result.vocab:
				raise DuplicateEntry(word)

			word_id = len(result.index2word)
			result.vocab[word] = Vocab(index=word_id, count=vocab_size - word_id)
			result.vectors[word_id] = weights
			result.index2word.append(word)

		real_size = len(result.index2word)

		if real_size != vocab_size:
			if discard is None:
				raise EOFError("unexpected end of input; is vocab_size incorrect or file otherwise damaged?")
			else:
				result.vectors.resize((real_size, vector_size)) # this should be no-copy

		return result

	def add(self, wwc):
		# type: (Collection[Tuple[str, np.ndarray, int]], ) -> None

		""" `wwc` is a collection of (word, weights, count) tuples """

		vocab_size, vector_size = self.vectors.shape
		self.vectors.resize((vocab_size + len(wwc), vector_size))
		for i, (word, weights, count) in enumerate(wwc):
			word_id = vocab_size + i
			voc = Vocab(index=word_id, count=count)
			self.vocab[word] = voc
			self.vectors[word_id] = weights
			self.index2word.append(word)
			yield voc

	def add_word(self, word, weights=None, count=1):
		if not weights:
			vocab_size, vector_size = self.vectors.shape
			weights = np.random.randn(1, vector_size)

		return list(self.add([(word, weights, count)]))[0]

	@classmethod
	def load_muse_format(cls, fname, limit=None, discard=None):
		# type: (Union[str, TextIO], int, Optional[Callable[[str], bool]]) -> KeyedVectorsOriginal

		""" reads at most `limit` vectors from muse file, if `discard` is not None it might be less.
			discard is a callable which should return False if word should not be loaded from file.
		"""

		with PathOrTextIO(fname, "rt", encoding="utf-8", newline="\n") as fin:
			vocab_size_, vector_size = map(int, next(fin).split(" "))
			if limit:
				vs = min(vocab_size_, limit)
			else:
				vs = vocab_size_

			try:
				return cls.load_a_format(islice(fin, limit), vs, vector_size, discard=discard)
			except DuplicateEntry as e:
				raise ValueError("duplicate word '{}' in {}".format(e, fname))

	@classmethod
	def load_glove_format(cls, fname, limit, vector_size, discard=None):
		# type: (Union[str, TextIO], int, int, Optional[Callable[[str], bool]]) -> KeyedVectorsOriginal

		""" Reads at most `limit` vectors from glove file, if `discard` is not None it might be less.
			`discard` is a callable which should return False if word should not be loaded from file.
		"""

		with PathOrTextIO(fname, "rt", encoding="utf-8") as fin:
			try:
				return cls.load_a_format(islice(fin, limit), limit, vector_size, discard=discard)
			except DuplicateEntry as e:
				raise ValueError("duplicate word '{}' in {}".format(e, fname))

	def get_keras_embedding(self, train_embeddings=False, mask_zero=True):
		try:
			from keras.layers import Embedding
		except ImportError:
			raise ImportError("Please install Keras to use this function")

		if mask_zero == True:
			self.index2word = [None]+ self.index2word
			zero_vec = np.zeros((1, self.vector_size))
			self.vectors = np.concatenate([zero_vec, self.vectors], axis=0)

		weights = self.vectors

		# set `trainable` as `False` to use the pretrained word embedding
		# No extra mem usage here as `Embedding` layer doesn't create any new matrix for weights
		layer = Embedding(
			input_dim=weights.shape[0], output_dim=weights.shape[1],
			weights=[weights], trainable=train_embeddings, mask_zero=mask_zero
		)
		return layer

	def transform_words_to_indices(self, words):
		# type: (Collection[str], ) -> Iterator[int]

		return (self.vocab[word].index for word in words)

	def transform_words_to_indices_tuple(self, words):
		# type: (Collection[str], ) -> Tuple[int, ...]

		return tuple(self.transform_words_to_indices(words))

if __name__ == "__main__":

	import argparse
	parser = argparse.ArgumentParser()
	parser.add_argument("glove_file")
	parser.add_argument("muse_file")
	args = parser.parse_args()

	word_vectors = KeyedVectors.load_glove_format(args.glove_file, 400000, 300)
	print(word_vectors.most_similar(positive=['woman', 'king'], negative=['man']))
	word_vectors = KeyedVectors.load_muse_format(args.muse_file)
	print(word_vectors.most_similar(positive=['woman', 'king'], negative=['man']))



from __future__ import absolute_import, division, print_function, unicode_literals

from builtins import range, input
from future.moves.itertools import zip_longest
from future.utils import PY2

from unittest import TestCase
from unittest.runner import TextTestRunner
from unittest.result import TestResult
from os import remove
from collections import defaultdict
from functools import wraps
from itertools import product
from tempfile import NamedTemporaryFile
from time import sleep
from io import TextIOWrapper
from typing import TYPE_CHECKING

from .ops import logical_xor, logical_implication
from .signal import HandleKeyboardInterrupt # problem here

if TYPE_CHECKING:
	from typing import Any, Iterable, Optional, Hashable, Tuple, Callable, DefaultDict

class nullcontext(object): # see: .compat.contextlib

	def __init__(self, *args, **kwargs):
		pass

	def __enter__(self):
		return self

	def __exit__(self, exc_type, exc_value, traceback):
		pass

anullcontext = nullcontext()

class MyTestResult(TestResult):
	"""Extension of TestResult to support numbering test cases"""

	def __init__(self, stream, descriptions, verbosity):
		"""Initializes the test number generator, then calls super impl"""
		self.stream = stream
		self.current_test = 0
		self.current_subtest = 0
		return TestResult.__init__(self, stream, descriptions, verbosity)

	def progress(self):
		self.stream.write("Running test #{}, subtest #{}\r".format(self.current_test, self.current_subtest))

	def startTest(self, test):
		self.current_test += 1
		TestResult.startTest(self, test)
		self.progress()

	def stopTest(self, test):
		TestResult.stopTest(self, test)

	def stopTestRun(self):
		self.stream.write("\n")
		TestResult.stopTestRun(self)

	def addSubTest(self, test, subtest, outcome):
		""" Called when a subtest finishes. """

		TestResult.addSubTest(self, test, subtest, outcome)
		self.testsRun += 1
		self.current_subtest += 1
		self.progress()

class MyTestRunner(TextTestRunner):
	"""Extension of TestRunner to support numbering test cases"""

	resultclass = MyTestResult

	def run(self, test):
		"""Stores the total count of test cases, then calls super impl"""

		total_test_cases = test.countTestCases()
		return TextTestRunner.run(self, test)

	def _makeResult(self):
		return TextTestRunner._makeResult(self)

class NoRaise(object):

	def __init__(self, testcase, message=None):
		self.testcase = testcase
		self.message = message

	def __enter__(self):
		pass

	def __exit__(self, exc_type, exc_value, traceback):
		if exc_type:
			if self.message:
				self.testcase.fail(self.message)
			else:
				self.testcase.fail(exc_value)

class MyTestCase(TestCase):

	def subTest(self, msg=None, **params):
		try:
			func = TestCase.subTest
		except AttributeError:
			return anullcontext
		else:
			return func(self, msg, **params)

	def assertAnd(self, first, second, msg=None):
		# type: (Any, Any, Optional[str]) -> None

		self.assertTrue(first and second, msg)

	def assertIterEqual(self, first, second, msg=None):
		# type: (Iterable, Iterable, Optional[str]) -> None

		for i, (a, b) in enumerate(zip_longest(first, second)):
			if msg:
				msg = " : " + str(msg)
			self.assertEqual(a, b, msg="in iteration index {}: {}".format(i, msg))

	def assertIterAlmostEqual(self, first, second, msg=None):
		# type: (Iterable, Iterable, Optional[str]) -> None

		for i, (a, b) in enumerate(zip_longest(first, second)):
			if msg:
				msg = " : " + str(msg)
			try:
				self.assertAlmostEqual(a, b, msg="in iteration index {}: {}".format(i, msg))
			except TypeError:
				# assert False, "Invalid types (probably different length iters)"
				raise AssertionError("Invalid types (probably different length iters)") # from None

	def assertAllEqual(self, args, msg=None): # *args doesn't work in python2!?
		# type: (Iterable, Optional[str]) -> None

		it = iter(args)
		first = next(it)
		for second in it:
			self.assertEqual(first, second, msg)

	def assertIterIterEqual(self, first, second, msg=None): # UNTESTED
		# type: (Iterable[Iterable], Iterable[Iterable], Optional[str]) -> None

		for i, (a, b) in enumerate(zip_longest(first, second)):
			if msg:
				msg = " : " + str(msg)
			self.assertIterEqual(a, b, msg="in iteration index {}: {}".format(i, msg))

	def assertPriorityEqual(self, first, second, msg=None):
		# type: (Iterable[Optional[Tuple[Hashable, Hashable]]], Iterable[Optional[Tuple[Hashable, Hashable]]], Optional[str]) -> None

		d1 = defaultdict(set) # type: DefaultDict[Hashable, set]
		d2 = defaultdict(set) # type: DefaultDict[Hashable, set]

		for a, b in zip_longest(first, second):
			if a is not None:
				val_1, p_1 = a
				if b is None and p_1 == 0:
					continue

			if b is not None:
				val_2, p_2 = b
				if a is None and p_2 == 0:
					continue

			self.assertAnd(a, b)

			if val_1 == 0 and val_2 == 0:
				continue

			d1[p_1].add(val_1)
			d2[p_2].add(val_2)
		self.assertEqual(d1, d2, msg)

	def assertUnorderedSeqEqual(self, first, second, msg=None):
		# type: (Iterable, Iterable, Optional[str]) -> None

		first = sorted(first)
		second = sorted(second)
		self.assertEqual(first, second, msg)

	def assertNoRaise(self, msg=None):
		return NoRaise(self, msg)

def random_arguments(n, *funcs):
	# type: (int, *Callable[[], Any]) -> Callable[[Callable], Callable]

	def decorator(func):
		@wraps(func)
		def inner(self):
			for i in range(n):
				with self.subTest(str(i)):
					assert func(self, *(f() for f in funcs)) is None
		return inner
	return decorator

# also called: parameterize
def parametrize(*args_list):
	# # type: (*tuple, ) -> Callable[[Callable], Callable] # mypy error: syntax error in type comment

	def decorator(func):
		@wraps(func)
		def inner(self):
			for args in args_list:
				with self.subTest(str(args)):
					assert func(self, *args) is None
		return inner
	return decorator

def parametrize_product(*args_list):
	# # type: (*tuple, ) -> Callable[[Callable], Callable] # mypy error: syntax error in type comment

	def decorator(func):
		@wraps(func)
		def inner(self):
			for args in product(*args_list):
				with self.subTest(str(args)):
					assert func(self, *args) is None
		return inner
	return decorator

def repeat(number):
	# type: (int, ) -> Callable[[Callable], Callable]

	def decorator(func):
		@wraps(func)
		def inner(self):
			for i in range(number):
				assert func(self) is None # no self.subTest(str(i))
		return inner
	return decorator

class closeable_tempfile(object):

	Uninterrupted = HandleKeyboardInterrupt()

	def __init__(self, mode="w+b", encoding=None):

		is_text = "t" in mode
		is_binary = "b" in mode

		assert logical_xor(is_text, is_binary), "Explicit text or binary mode required: {}".format(mode)
		assert logical_implication(is_binary, encoding is None)

		if is_text:
			encoding = encoding or "utf-8"

		if PY2:
			ntf = NamedTemporaryFile(mode=mode, delete=False)
			if encoding:
				ntf.readable = lambda: True
				ntf.writable = lambda: True
				ntf.seekable = lambda: True

				self.f = TextIOWrapper(ntf, encoding)
			else:
				self.f = ntf
		else:
			self.f = NamedTemporaryFile(mode=mode, encoding=encoding, delete=False)

	def __enter__(self):
		return self.f, self.f.name

	def __exit__2(self, type, value, traceback):
		with self.Uninterrupted:
			self.f.close() # close in case user has not closed the file
			remove(self.f.name)

	def __exit__(self, type, value, traceback):
		try:
			if type == EOFError:
				sleep(1) # wait for KeyboardInterrupt which might follow EOFError...
		finally:
			self.f.close() # close in case user has not closed the file
			remove(self.f.name)



from __future__ import absolute_import, division, print_function, unicode_literals

from typing import TYPE_CHECKING

if TYPE_CHECKING:
	from typing import MutableSequence, Callable

from past.builtins import cmp

def _insertion(seq, cmp_, left, right, gap):
	# type: (MutableSequence, Callable, int, int, int)

	# print("_insertion", left, right, gap)

	loc = left+gap
	while loc <= right:
		i = loc - gap
		value = seq[loc]
		while i >= left and cmp_(seq[i], value) > 0:
			seq[i+gap] = seq[i]
			i -= gap
		seq[i+gap] = value
		loc += gap

GROUP_SIZE = 5

def median_of_medians(seq, cmp_=None, left=None, right=None, gap=1):
	# type: (MutableSequence, Callable, int, int, int)

	""" Approximate median selection algorithm. """

	cmp_ = cmp_ or cmp

	if not left:
		left = 0
	if not right:
		right = len(seq) - 1

	span = GROUP_SIZE*gap
	num = (right - left + 1) // span

	# print("median_of_medians", left, right, gap, span, num)

	if num == 0:
		_insertion(seq, cmp_, left, right, gap)
		num = (right - left + 1) // gap
		return left + gap*(num-1) // 2

	s = left
	while s < right-span:
		_insertion(seq, cmp_, s, s + span - 1, gap)
		s += span

	if num < GROUP_SIZE:
		_insertion(seq, cmp_, left+span // 2, right, span)
		return left + num*span // 2
	else:
		return median_of_medians(seq, cmp_, left + span // 2, s-1, span)



import logging
from functools import wraps
from collections import defaultdict
from typing import TYPE_CHECKING

if TYPE_CHECKING:
	from typing import Callable, Collection

def _to_union(types):
	# type: (Collection[str], ) -> str

	types = list(set(types))

	if not types:
		return ""
	elif len(types) == 1:
		return types[0]
	elif len(types) > 1:
		return "Union[" + ", ".join(types) + "]"

def _type_str(obj):
	# type: (Any, ) -> str

	return type(obj).__name__

def rec_repr(obj):
	# type: (Any, ) -> str

	if isinstance(obj, defaultdict):
		return _type_str(obj) + "[" + rec_repr(obj.default_factory) + "]"
	elif isinstance(obj, list):
		return "List[" + _to_union(map(rec_repr, obj)) + "]"
	elif isinstance(obj, set):
		return "Set[" + _to_union(map(rec_repr, obj)) + "]"
	else:
		return _type_str(obj)

def log_wrap_call(func):
	# type: (Callable, ) -> Callable

	""" Decorator which logs all calls to function with all arguments """

	@wraps(func)
	def inner(*args, **kwargs):
		logging.debug("START '{}' with {}, {}".format(func.__name__, str(args), str(kwargs)))

		try:
			ret = func(*args, **kwargs)
		except BaseException as e:
			logging.debug("'{}' RAISED '{}' with {}, {}".format(func.__name__, str(e), str(args), str(kwargs)))
			raise

		logging.debug("END '{}' with {}, {}".format(func.__name__, str(args), str(kwargs)))
		return ret
	return inner



from __future__ import absolute_import, division, print_function, unicode_literals
from builtins import input

import os, sys, os.path, argparse, shlex
from pathlib import Path
from functools import wraps

from .iter import is_empty
from .stdio import confirm

def get_args(argparser):
	""" get commandline arguments from std input instead """

	from pprint import pprint

	if len(sys.argv) > 1:
		return argparser.parse_args()

	print("stdin")
	args = []

	for action in argparser._actions:
		pprint(action)

	for k, v in argparser._option_string_actions.items():

		nargs = v.nargs if v.nargs is not None else 1

		if nargs == 0:
			if confirm(str(k)):
				args.append(k)
		else:
			instr = input("{} ({}): ".format(k, nargs)) # separate multiple value by whitespace, quoting supported.
			args.append(k)
			args.append(instr)

	args = shlex.split(" ".join(args))
	pprint(args)
	return argparser.parse_args(args)

def arg_to_path(func):
	# type: (Callable, ) -> Callable

	@wraps(func)
	def inner(path):
		return func(Path(path))
	return inner

@arg_to_path
def is_dir(path):
	# type: (Path, ) -> Path

	"""Checks if a path is an actual directory"""

	if not path.is_dir():
		msg = "{0} is not a directory".format(path)
		raise argparse.ArgumentTypeError(msg)

	return path

@arg_to_path
def is_file(path):
	# type: (Path, ) -> Path

	"""Checks if a path is an actual file"""

	if not path.is_file():
		msg = "{0} is not a file".format(path)
		raise argparse.ArgumentTypeError(msg)

	return path

@arg_to_path
def future_file(path):
	# type: (Path, ) -> Path

	""" Tests if file can be created to catch errors early.
		Checks if directory is writeable and file does not exist yet.
	"""

	if path.parent and not os.access(str(path.parent), os.W_OK):
		msg = "cannot access directory {0}".format(path.parent)
		raise argparse.ArgumentTypeError(msg)
	if path.is_file():
		msg = "file {0} already exists".format(path)
		raise argparse.ArgumentTypeError(msg)
	return path

@arg_to_path
def empty_dir(dirname):
	# type: (Path, ) -> Path

	""" tests if directory is empty """

	if not is_empty(os.scandir(dirname)):
		msg = "directory {0} is not empty".format(dirname)
		raise argparse.ArgumentTypeError(msg)

	return dirname

if __name__ == "__main__":
	parser = argparse.ArgumentParser()
	parser.add_argument('--str')
	parser.add_argument('--required', action='store_true')
	parser.add_argument('--add', nargs="+")
	print(get_args(parser))

	"""
	parser = argparse.ArgumentParser()
	parser.add_argument('--indir', type=is_dir)
	parser.add_argument('--infile', type=is_file)
	parser.add_argument('--outfile', type=future_file)
	parser.add_argument('--outdir', type=empty_dir)
	args = parser.parse_args()
	print(args)
	"""



from __future__ import absolute_import, division, print_function, unicode_literals

from .compat.datetime import datetime, timezone

def now():
	# type: () -> datetime

	""" Returns the current datetime as timezone aware object in UTC timezone. """

	return datetime.now(timezone.utc)



from __future__ import absolute_import, division, print_function, unicode_literals

import pickle
from functools import wraps
from typing import TYPE_CHECKING

from .twothree import FileNotFoundError
from .file import copen, OpenFileAndDeleteOnError

if TYPE_CHECKING:
	from typing import Any, Callable

def cache(path, protocol=None):
	# type: (str, int) -> Callable[[Callable], Callable]

	""" Decorator to cache function calls. Doesn't take function arguments into regard. """

	def decorator(func):
		# type: (Callable, ) -> Callable

		@wraps(func)
		def inner(*args, **kwargs):
			# type: (*Any, **Any) -> Any

			try:
				with copen(path, "rb") as fr:
					return pickle.load(fr)
			except FileNotFoundError:
				result = func(*args, **kwargs)
				with OpenFileAndDeleteOnError(path, "wb") as fw:
					# if pickling fails, the file will be deleted
					# and no incomplete file will be left on the disk
					pickle.dump(result, fw, protocol=protocol)
				return result
		return inner
	return decorator



from __future__ import absolute_import, division, print_function, unicode_literals

import signal, logging, sys
from typing import TYPE_CHECKING

if TYPE_CHECKING:
	from typing import TypeVar, Any, Iterable, Callable, Union
	from types import FrameType
	if sys.version_info >= (3, 5):
		from signal import Signals # pylint: disable=no-name-in-module
	else:
		Signals = int

	T = TypeVar("T")

logger = logging.getLogger(__name__)

class HandleKeyboardInterrupt(object):

	""" Context manager to handle ctrl-c (keyboard interrupts)
	
		with HandleKeyboardInterrupt():
			do_a()
			do_b()

		Any KeyboardInterrupt received within the context manager will be caught and possibly
		re-raised at the end of the block. It can be used to make sure that sensitive code is not
		carelessly interrupted in the middle of execution by the user. The process could still be
		kill though, so this method should only be used as a convenience function, not to implement
		atomic transactions.

		https://stackoverflow.com/questions/842557/how-to-prevent-a-block-of-code-from-being-interrupted-by-keyboardinterrupt-in-py#comment81343976_21919644
		"this code may call third-party exception handlers in threads other than the main thread, which CPython never does"
		so be careful with threads
	"""

	def __init__(self, raise_after=True, delay=-1):
		# type: (bool, int) -> None

		""" If `raise_after` is True, the KeyboardInterrupt will be re-raised when leaving the context.
			`delay` specifies the maximum number of times the signal is caught.
		"""

		self.init_raise_after = raise_after
		self.init_delay = delay

	def __enter__(self):
		# type: () -> None

		self.raise_after = self.init_raise_after # pylint: disable=attribute-defined-outside-init
		self.delay = self.init_delay # pylint: disable=attribute-defined-outside-init
		self.signal_received = None # type: Optional[tuple]
		self.old_handler = signal.signal(signal.SIGINT, self.handler) # type: ignore # cast to Callable[[Signals, FrameType], None]

	def handler(self, sig, frame):
		# type: (Signals, FrameType) -> None

		if self.delay == 0:
			self.raise_after = False
			self.old_handler(sig, frame) # type: ignore # see cast above
		else:
			self.delay -= 1
			self.signal_received = (sig, frame)
			logger.debug("SIGINT received. Delaying KeyboardInterrupt.")

	def __exit__(self, type, value, traceback):
		# should this show a warning for 'IOError: [Errno 4] Interrupted function call' on python 2.7?

		signal.signal(signal.SIGINT, self.old_handler)
		if self.signal_received and self.raise_after:
			self.old_handler(*self.signal_received)

def safe_for_loop(it, body_func, else_func=None):
	# type: (Iterable[T], Callable[[T], bool], Callable[[], Any]) -> None

	""" This function makes sure that `body_func(element)` is called each time an element is retrieved from the
		iterable `it`, even if a `KeyboardInterrupt` is signaled in-between.
		After `body_func` is called, the KeyboardInterrupt will be re-raised and can be caught in
		calling function. `body_func` can optionally return `True` to exit the loop early.
		If the loop finishes, `else_func` will be called if provided.
	"""

	it = iter(it)
	finished = False
	Uninterrupted = HandleKeyboardInterrupt(True)
	while True:
		with Uninterrupted:
			try:
				chunk = next(it)
			except StopIteration:
				finished = True
				break
			if body_func(chunk):
				break
	if finished and else_func: # no break
		with Uninterrupted:
			else_func()



from __future__ import absolute_import, division, print_function, unicode_literals

from builtins import map
from typing import TYPE_CHECKING

import numpy as np

from .iter import pairwise
from .cv import iter_video
from .image import image_histogram, image_block_histogram
from .numpy import histogram_correlation

if TYPE_CHECKING:
	pass

def scene_change_detection_histogram_correlation(images):

	""" Based on: 'Histogram Correlation for Video Scene Change Detection'. Algorithm described
		in paper is incomplete. It only mentions 'almost constant', 'not constant' and 'change is sharp'
		behaviour of the correlation. A real implementation would need to define a sliding window
		based derivative and threshold values for said derivatives.
	"""

	it = iter(images)
	hist_rf = image_histogram(np.array(next(it)))

	for image in it:
		hist_i = image_histogram(np.array(image))
		yield histogram_correlation(hist_rf, hist_i)

def scene_change_detection_block_histogram(images):
	it = iter(images)

	def get_means_of_block_histograms(arr):
		hist = image_block_histogram(arr, 16, 16)
		#print(hist.shape)
		#print(hist.tolist())
		means = np.mean(hist, axis=-1)
		print(means)
		return means

	for means1, means2 in pairwise(map(get_means_of_block_histograms, it)):
		# difference of means
		diffs = np.abs(means1 - means2)
		yield diffs

def proc(hist, lambda_=200):
	delta = 0.5
	DS = np.sqrt(2.)/2.*np.sum(np.abs(hist - hist.T), axis=(-2,-1))
	B = DS > lambda_
	BC = np.mean(unpack(B, 2, 2), axis=-1) > delta # noqa: F821

def test_scene_change_detection_histogram_correlation():
	images = iter_video()
	for score in scene_change_detection_histogram_correlation(images):
		print(score)

def test_scene_change_detection_block_histogram():
	images = iter_video()
	for diffs in scene_change_detection_block_histogram(images):
		print(diffs.tolist())

if __name__ == "__main__":
	from PIL import Image
	#test_image_histogram_gray()
	#test_scene_change_detection_histogram_correlation()
	test_scene_change_detection_block_histogram()



from __future__ import absolute_import, division, print_function, unicode_literals

from io import open
from pathlib import Path
try:
	from contextlib import ExitStack
except ImportError:
	from contextlib2 import ExitStack

from PyPDF2 import PdfFileMerger

from .twothree import FileNotFoundError

def join_pdfs_in_folder(path_in, file_out, overwrite=False):
	# type: (Union[Path, str], Union[Path, str]) -> None

	""" Join all PDFs in `path_in` and write to `file_out`. """

	if not isinstance(path_in, Path):
		path_in = Path(path_in)

	merger = PdfFileMerger(strict=True)

	if not path_in.exists():
		raise FileNotFoundError(path_in)

	with ExitStack() as stack:
		for path in path_in.glob("*.pdf"):
			if path.is_file():
				fr = stack.enter_context(open(path, "rb"))
				merger.append(fr)

		if overwrite:
			mode = "w"
		else:
			mode = "x"

		with open(file_out, mode+"b") as fw:
			merger.write(fw)

if __name__ == "__main__":
	from argparse import ArgumentParser

	parser = ArgumentParser(description='Merge pdf files in directory into one file.')
	parser.add_argument('dir', help='input directory')
	parser.add_argument('out', help='output file path')
	args = parser.parse_args()

	join_pdfs_in_folder(args.dir, args.out)



from __future__ import absolute_import, division, print_function, unicode_literals

from typing import TYPE_CHECKING

import cv2

if TYPE_CHECKING:
	from typing import Union, Iterator
	import numpy as np

def iter_video(input=0, show=False):
	# type: (Union[str, int], bool) -> Iterator[np.ndarray]

	cap = cv2.VideoCapture(input)

	try:
		while True:
			retval, image = cap.read()
			if retval:
				yield image
			else:
				break
			
			if show:
				cv2.imshow('iter_video', image)
				if cv2.waitKey(1) & 0xFF == ord('q'):
					return
	finally:
		cap.release()
		if show:
			cv2.destroyAllWindows()



from __future__ import absolute_import, division, print_function, unicode_literals
from builtins import chr
from future.moves.itertools import zip_longest

import os, stat, os.path, re, logging
from operator import attrgetter
from fnmatch import fnmatch
from typing import TYPE_CHECKING

try:
	from os import scandir
except ImportError:
	from scandir import scandir # backport

from .file import equal_files
from .iter import is_empty
from .os import islink, uncabspath
from .ops import logical_implication
from .string import replace_list

if __debug__:
	import unidecode

if TYPE_CHECKING:
	from typing import Union, Iterable, Iterator
	from pathlib import Path

	MyDirEntryT = Union[os.DirEntry, SkippableDirEntry]

logger = logging.getLogger(__name__)

ascii_control = set(chr(i) for i in range(1, 32))
ntfs_illegal_chars = {"\0", "/"}
fat_illegal_chars = {"\0", "/", '?', '<', '>', '\\', ':', '*', '|', '"', '^'} # ^ really?

windows_ntfs_illegal_chars = ntfs_illegal_chars | {"\\", ":", "*", "?", "\"", "<", ">", "|"} | ascii_control
windows_fat_illegal_chars = fat_illegal_chars

windows_illegal_chars = windows_ntfs_illegal_chars | windows_fat_illegal_chars
unix_illegal_chars = {"\0", "/"}
mac_illegal_chars = {"\0", "/", ":"}

windows_reserved_names = {"CON", "PRN", "AUX", "NUL", "COM1", "COM2", "COM3", "COM4", "COM5", "COM6", "COM7", "COM8", "COM9", "LPT1", "LPT2", "LPT3", "LPT4", "LPT5", "LPT6", "LPT7", "LPT8", "LPT9"}

class fileextensions(object):
	audio = ("wav", "mp3", "aac", "flac", "m4a", "m4b", "aiff", "ogg", "wma", "mka", "ac3", "dts")
	video = ("avi", "mp4", "m4v", "mkv", "mpg", "mpeg", "wmv", "mov", "flv", "f4v", "webm", "vob", "ogv", "m2ts", "rm", "rmvb", "asf", "3gp", "nsv", "divx", "ogm")
	images = ("bmp", "jpg", "jpeg", "png", "gif", "tif", "tiff", "tga")

	archives = ("zip", "rar", "ace", "tar", "tgz", "tbz", "7z", "cab", "dmg", "wim")
	executables = ("exe", "dll", "so", "scr", "msi", "pyd", "sys")
	scripts = ("bat", "ps1", "sh", "py", "php")
	documents = ("txt", "pdf", "doc", "docx", "ps")
	image_archives = ("cbz", "cbr", "cb7", "cbt", "cba")

	disc_images_binary = ("iso", "bin", "mdf", "img")
	disc_images_sidecar = ("cue", "mds", "ccd")
	disc_images = disc_images_binary + disc_images_sidecar

	game_images = ("gcn", "nds", "3ds", "wad", "xbx", "gba", "gb", "nes", "sfc", "n64", "z64")
	compressed = ("gz", "bz2", "lzma", "xz", "z")
	subtitles = ("srt", "sub", "idx", "sup", "ssa", "ass")

	scene = ("nfo", "sfv")

class IllegalFilename(ValueError):
	pass

class WindowsIllegalFilename(IllegalFilename):
	pass

class FileProperties(object):

	__slots__ = ("relpath", "size", "isdir", "abspath", "id", "modtime", "hash")

	def __init__(self, relpath, size, isdir, abspath=None, id=None, modtime=None, hash=None):
		self.relpath = relpath
		self.size = size
		self.isdir = isdir
		self.abspath = abspath
		self.id = id
		self.modtime = modtime
		self.hash = hash

	def __iter__(self):
		return (self.relpath, self.size, self.isdir, self.abspath, self.id, self.modtime, self.hash)

	def __repr__(self):
		keys = self.__slots__
		values = attrgetter(*keys)(self)
		args = ("{}={!r}".format(k, v) for k, v in zip(keys, values) if v is not None)
		return "FileProperties({})".format(", ".join(args))

class DirEntryStub(object):
	__slots__ = ('name', 'path')

	def __init__(self, name, path):
		self.name = name
		self.path = path

class SkippableDirEntry(object):
	__slots__ = ('entry', '_path', 'follow')

	def __init__(self, entry):
		self.entry = entry
		self._path = entry.path
		self.follow = True

	def stat(self):
		return self.entry.stat()

	def is_dir(self):
		return self.entry.is_dir()

	def is_file(self):
		return self.entry.is_file()

	def is_symlink(self):
		return self.entry.is_symlink()

	def inode(self):
		return self.entry.inode()

	@property
	def name(self):
		return self.entry.name

	@property
	def path(self):
		return self._path

	@path.setter
	def path(self, path):
		self._path = path

def st_mode_to_str(st_mode):
	# type: (int, ) -> str

	if stat.S_ISDIR(st_mode): return "directory"
	elif stat.S_ISREG(st_mode): return "regular file"
	elif stat.S_ISCHR(st_mode): return "character special device file"
	elif stat.S_ISBLK(st_mode): return "block special device file"
	elif stat.S_ISFIFO(st_mode): return "named pipe"
	elif stat.S_ISLNK(st_mode): return "symbolic link"
	elif stat.S_ISSOCK(st_mode): return "socket"
	else: return "unknown"

def append_to_filename(path, s):

	""" Add `s` to the filename given in `path`. It's added before the extension, not after. """

	root, ext = os.path.splitext(path)
	return root + s + ext

def scandir_error_log(entry, exception):
	logger.exception("Error in %s", entry, exc_info=exception)

def scandir_error_raise(entry, exception):
	raise exception

def scandir_error_ignore(entry, exception):
	pass

def _scandir_rec_skippable(rootentry, files=True, others=False, follow_symlinks=True, errorfunc=scandir_error_raise):
	# type: (DirEntry, bool, bool, bool, Callable[[DirEntry, Exception], None]) -> Iterator[SkippableDirEntry]

	try:
		with scandir(rootentry.path) as it: # python 3.6
			for entry in it:
				if files and entry.is_file(follow_symlinks=follow_symlinks):
					yield SkippableDirEntry(entry)
				elif entry.is_dir(follow_symlinks=follow_symlinks):
					if not follow_symlinks and islink(entry): # must be a windows junction
						continue
					se = SkippableDirEntry(entry)
					yield se
					if se.follow:
						for e in _scandir_rec_skippable(entry, files, others, follow_symlinks, errorfunc):
							yield e
				else:
					if others:
						yield SkippableDirEntry(entry)
	except OSError as e:
		errorfunc(rootentry, e)

def _scandir_rec(rootentry, files=True, dirs=False, others=False, rec=True, follow_symlinks=True, errorfunc=scandir_error_raise):
	# type: (DirEntry, bool, bool, bool, bool, bool, Callable[[DirEntry, Exception], None]) -> Iterator[DirEntry]

	try:
		with scandir(rootentry.path) as it: # python 3.6
			for entry in it:
				if files and entry.is_file(follow_symlinks=follow_symlinks):
					yield entry
				elif entry.is_dir(follow_symlinks=follow_symlinks):
					if not follow_symlinks and islink(entry): # must be a windows junction
						continue
					if dirs:
						yield entry
					if rec:
						for e in _scandir_rec(entry, files, dirs, others, rec, follow_symlinks, errorfunc):
							yield e
				else:
					if others:
						yield entry
	except OSError as e:
		errorfunc(rootentry, e)

def scandir_rec(path, files=True, dirs=False, others=False, rec=True, follow_symlinks=True, relative=False, allow_skip=False, errorfunc=scandir_error_log):
	# type: (str, bool, bool, bool, bool, bool, bool, bool, Callable[[MyDirEntryT, Exception], None]) -> Iterator[MyDirEntryT]

	assert logical_implication(allow_skip, dirs and rec)

	entry = DirEntryStub(os.path.basename(path), uncabspath(path)) # for python 2 compat. and long filename support
	if not allow_skip:
		it = _scandir_rec(entry, files, dirs, others, rec, follow_symlinks, errorfunc)
	else:
		it = _scandir_rec_skippable(entry, files, others, follow_symlinks, errorfunc)

	if not relative:
		return it
	else:
		basepath = uncabspath(path)
		if not allow_skip:
			def modpathrelative(entry):
				entry = SkippableDirEntry(entry)
				entry.path = os.path.relpath(entry.path, basepath)
				return entry
		else:
			def modpathrelative(entry):
				entry.path = os.path.relpath(entry.path, basepath)
				return entry

		return map(modpathrelative, it)

def join_ext(name, ext):
	# type: (str, str) -> str

	""" adds extenions to name, eg. `join_ext("picture", "jpg") -> "picture.jpg" """

	return name + "." + ext

def make_readonly(path, stats=None):
	# type: (Path, Optional[os.stat_result]) -> None

	"""deletes all write flags"""
	if not stats:
		stats = os.stat(path)
	os.chmod(path, stat.S_IMODE(stats.st_mode) & ~stat.S_IWUSR & ~stat.S_IWGRP & ~stat.S_IWOTH)

def make_writeable(path, stats=None):
	# type: (Path, Optional[os.stat_result]) -> None

	"""set owner write flag"""
	if not stats:
		stats = os.stat(path)
	os.chmod(path, stat.S_IMODE(stats.st_mode) | stat.S_IWRITE)

def is_writeable(stats):
	# type: (os.stat_result, ) -> bool
	return stats.st_mode & stat.S_IWRITE

def is_readable(stats):
	# type: (os.stat_result, ) -> bool
	return stats.st_mode & stat.S_IREAD

def isfile(stats):
	# type: (os.stat_result, ) -> bool
	return stat.S_ISREG(stats.st_mode)

def isdir(stats):
	# type: (os.stat_result, ) -> bool
	return stat.S_ISDIR(stats.st_mode)

# was: islink
def statsislink(stats):
	# type: (os.stat_result, ) -> bool
	return stat.S_ISLNK(stats.st_mode)

def extract_basic_stat_info(stats):
	# type: (os.stat_result, ) -> tuple

	""" Returns size, modification, creation time and access mode. """
	return (stats.st_size, stats.st_mtime, stats.st_ctime, stats.st_mode)

def safe_filename(filename, replace=""):
	# type: (str, str) -> str

	WIN = windows_illegal_chars
	UNIX = unix_illegal_chars
	MAC = mac_illegal_chars

	return replace_list(filename, set.union(WIN, UNIX, MAC), replace) # use translate or something similar fast

safe_filename_simple = safe_filename

def windows_compliant_filename(filename, replace="_"):
	# type: (str, str) -> str
	""" https://msdn.microsoft.com/en-us/library/aa365247.aspx """
	# reg. wikipedia: \x7f is valid!
	# trailing dots and spaces are ignored by windows

	if replace in windows_illegal_chars:
		raise ValueError("replace character cannot be a illegal character")

	ret = replace_list(filename, windows_illegal_chars, replace).rstrip(". ") # strip or replace?, translate

	root, _ = os.path.splitext(ret)

	if root.upper() in windows_reserved_names:
		raise WindowsIllegalFilename("filename would result in reserved name")

	return ret

def equal_dirs_iter(dir1, dir2, follow_symlinks=False):

	""" Tests if two directories are equal. Doesn't handle links. """

	def entry_path_size(entry):
		return entry.path, entry.stat().st_size

	files1 = sorted(map(entry_path_size, scandir_rec(dir1, follow_symlinks=follow_symlinks))) # get rel paths
	files2 = sorted(map(entry_path_size, scandir_rec(dir2, follow_symlinks=follow_symlinks))) # get rel paths

	for (path1, size1), (path2, size2) in zip_longest(files1, files2):
		if path1 != path2:
			yield ("name", path1, path2)

		elif size1 != size2:
			yield ("size", path1, path2)

	for (path1, size1), (path2, size2) in zip_longest(files1, files2):
		if path1 == path2 and size1 == size2 and not equal_files(path1, path2):
			yield ("data", path1, path2)

def equal_dirs(dir1, dir2):
	return is_empty(equal_dirs_iter(dir1, dir2))

def realpath_win(path):
	# type: (Path, ) -> str
	"""fix for os.path.realpath() which doesn't work under Win7"""
	# fixme: handle junctions

	path = os.path.abspath(path)
	if os.path.islink(path):
		return os.path.normpath(os.path.join(os.path.dirname(path), os.readlink(path)))
	else:
		return path

def convert_filenames_to_ascii(path, follow_symlinks=False):
	# type: (Path, ) -> None
	""" convert all files in `path` to a ascii representation using unidecode """

	from unidecode import unidecode

	for entry in scandir_rec(path, files=True, dirs=False, rec=False, follow_symlinks=follow_symlinks):
		filepath = entry.path
		base, name = os.path.split(filepath)
		os.rename(filepath, os.path.join(base, unidecode(name)))

def search(directories, pattern, dirs=True, files=True, rec=True, follow_symlinks=False):
	# type: (Iterable[Path], str, bool, bool, bool) -> Iterator[DirEntry]

	""" Search for files and folders matching the wildcard `pattern`. """

	for directory in directories:
		for entry in scandir_rec(directory, dirs=dirs, files=files, rec=rec, follow_symlinks=follow_symlinks):
			if fnmatch(entry.name, pattern):
				yield entry

def rename_files_in_folder(path, pattern, transform, template, rec=True, follow_symlinks=False):
	cpattern = re.compile(pattern)

	for entry in scandir_rec(path, files=True, dirs=False, rec=rec, follow_symlinks=follow_symlinks):
		filepath = entry.path
		base, name = os.path.split(filepath)

		m = cpattern.match(name)
		if not m:
			continue
		args = transform(*m.groups())

		to = template.format(*args)
		os.rename(filepath, os.path.join(base, to))

def clean_directory_by_extension(path, ext, rec=True, follow_symlinks=False):
	# type: (str, str, bool, bool) -> None

	""" Deletes all files of type `ext` if the same file without the ext exists
		and rename if it does not exit. """

	#delete .ext if normal file exists
	for entry in scandir_rec(path, files=True, dirs=False, rec=rec, follow_symlinks=follow_symlinks):
		filepath = entry.path
		if not filepath.endswith(ext):
			file_to_delete = filepath + ext
			if os.path.isfile(file_to_delete):
				logger.info("Remove: %s", file_to_delete)
				try:
					os.remove(file_to_delete)
				except OSError as e:
					logger.exception("Could not delete: %s", file_to_delete)
					break

	#rename all .ext to normal
	for entry in scandir_rec(path, files=True, dirs=False, rec=rec, follow_symlinks=follow_symlinks):
		filepath = entry.path
		if entry.name.endswith(ext):
			target = filepath[:-len(ext)]
			try:
				logger.info("Rename: %s -> %s", filepath, target)
				os.rename(filepath, target)
			except OSError as e:
				logger.exception("Should not have happened")
				break



from __future__ import absolute_import, division, print_function, unicode_literals

from builtins import range
from copy import deepcopy

from .compat.math import isqrt
from .math import row_indices, col_indices, subblock_indices
from .set import get as setget

class Unsolvable(Exception):
	pass

sudokus = {
	"very_easy": [
		5,4,2,0,1,7,0,9,8,
		9,6,8,5,4,2,0,0,7,
		7,0,1,6,9,8,2,5,0,
		8,5,7,4,2,9,1,6,0,
		1,2,3,8,7,6,9,0,5,
		6,9,4,1,0,3,0,7,2,
		4,1,0,2,0,5,7,0,9,
		3,8,0,7,6,4,5,2,0,
		2,0,5,9,3,1,4,8,6
	],

	"easy": [
		4,8,5,1,0,6,9,0,0,
		6,2,0,0,9,0,8,0,0,
		0,1,0,2,0,8,0,7,0,
		8,6,4,3,0,0,7,0,9,
		2,0,0,9,0,5,0,0,4,
		5,9,1,4,0,7,0,2,8,
		0,4,8,0,5,9,2,6,3,
		0,5,2,0,3,4,1,8,7,
		0,0,6,8,0,1,4,0,5
	],

	"normal": [
		5,0,1,3,4,2,0,0,0,
		0,0,2,0,1,0,0,5,3,
		3,0,7,5,0,0,0,0,0,
		9,0,8,0,0,3,0,4,1,
		0,0,0,0,9,0,0,0,0,
		1,4,0,2,7,0,5,0,6,
		0,0,0,0,2,4,8,0,9,
		7,1,4,0,6,0,3,0,0,
		0,0,0,0,3,0,0,0,4
	],

	"hard": [
		0,2,0,0,6,5,0,0,3,
		0,0,3,0,8,0,2,0,1,
		4,0,0,0,0,0,0,0,0,
		0,9,0,0,1,6,4,7,0,
		0,0,1,0,0,0,6,0,0,
		0,4,5,7,3,0,0,9,0,
		0,0,0,0,0,0,0,0,4,
		2,0,8,0,4,0,3,0,0,
		1,0,0,5,9,0,0,2,0
	],

	"very": [
		2,8,5,0,0,0,0,0,0,
		0,7,0,0,2,5,0,0,9,
		0,0,0,0,0,0,0,0,4,
		1,9,0,0,0,0,0,0,0,
		6,0,0,0,9,0,0,1,0,
		0,0,0,7,0,4,0,8,0,
		0,0,0,8,0,3,0,0,0,
		0,0,3,6,0,0,0,4,0,
		0,0,0,0,0,0,5,0,0
	],

	"suck": [
		0,0,5,3,0,0,0,0,0,
		8,0,0,0,0,0,0,2,0,
		0,7,0,0,1,0,5,0,0,
		4,0,0,0,0,5,3,0,0,
		0,1,0,0,7,0,0,0,6,
		0,0,3,2,0,0,0,8,0,
		0,6,0,5,0,0,0,0,9,
		0,0,4,0,0,0,0,3,0,
		0,0,0,0,0,9,7,0,0
	],

	"Arto_Inkala": [
		8,0,0,0,0,0,0,0,0,
		0,0,3,6,0,0,0,0,0,
		0,7,0,0,9,0,2,0,0,
		0,5,0,0,0,7,0,0,0,
		0,0,0,0,4,5,7,0,0,
		0,0,0,1,0,0,0,3,0,
		0,0,1,0,0,0,0,6,8,
		0,0,8,5,0,0,0,1,0,
		0,9,0,0,0,0,4,0,0
	],
	
	"norvig": [# unsolvable, doesn't have a solution
		0,0,0,0,0,5,0,8,0,
		0,0,0,6,0,1,0,4,3,
		0,0,0,0,0,0,0,0,0,
		0,1,0,5,0,0,0,0,0,
		0,0,0,1,0,6,0,0,0,
		3,0,0,0,0,0,0,0,5,
		5,3,0,0,0,0,0,6,1,
		0,0,0,0,0,0,0,0,4,
		0,0,0,0,0,0,0,0,0
	],
}

class Sudoku(object):

	def __init__(self, outer_square_size, square, sym_set, sym_free):
		assert isinstance(sym_set, set)

		self.outer_square_size = outer_square_size
		self.outer_square_area = self.outer_square_size * self.outer_square_size
		self.inner_square_size = isqrt(outer_square_size)
		if self.inner_square_size ** 2 != self.outer_square_size:
			raise ArithmeticError("outer_square_size is not a multiple of inner_square_size")
		self.square = square
		if len(self.square) != self.outer_square_area:
			raise ArithmeticError("elements in list != sudoku size")

		self.solved = False
		self.sym_set = sym_set
		self.sym_free = sym_free

	def init_square(self, sudoku):
		raise NotImplementedError

	def get_square(self):
		raise NotImplementedError

	def print_square(self):
		for i, num in enumerate(self.get_square(), 1):
			print(num, end=' ')
			if i % self.outer_square_size == 0:
				print()
		print()

	def solve(self):
		self.square = self.init_square(self.square)
		return self._solve()

class SudokuRulebased(Sudoku):

	def init_square(self, sudoku):
		square = []
		for i in sudoku:
			if i != self.sym_free:
				square.append({i})
			else:
				square.append(self.sym_set.copy())
		return square

	def get_square(self):
		for num in self.square:
			if len(num) == 1:
				yield setget(num)
			else:
				yield num

	def get_row_nums(self, i):
		row_set = set()
		for x in row_indices(i, self.outer_square_size):
			if len(self.square[x]) == 1 and x != i:
				row_set.update(self.square[x])
		return row_set

	def get_column_nums(self, i):
		column_set = set()
		for x in col_indices(i, self.outer_square_size):
			if len(self.square[x]) == 1 and x != i:
				column_set.update(self.square[x])
		return column_set

	def get_inner_square_nums(self, n):
		inner_square_set = set()
		for k in subblock_indices(n, self.outer_square_size, self.inner_square_size):
			if len(self.square[k]) == 1 and k != n:
				inner_square_set.update(self.square[k])
		return inner_square_set

	def update_cell(self, i):
		a = self.get_row_nums(i)
		b = self.get_column_nums(i)
		c = self.get_inner_square_nums(i)
		self.square[i] -= a
		self.square[i] -= b
		self.square[i] -= c

	def _solve(self):
		while not self.solved:
			self.solved = True
			square_old = deepcopy(self.square)
			for i in range(self.outer_square_area):
				if len(self.square[i]) != 1:
					self.solved = False
					#print_confirm("#%u is unsolved -> solve" % i)
					self.update_cell(i)
					#print_confirm(self.square)
			if square_old == self.square and not self.solved:
				raise NotImplementedError("Sudoku not solvable using this method")
		return self.square

class SudokuBruteforce(Sudoku):

	def init_square(self, sudoku):
		return sudoku

	def get_square(self):
		return self.square

	def get_row_nums(self, i):
		row_set = set()
		for x in row_indices(i, self.outer_square_size):
			if x != i:
				row_set.add(self.square[x])
		return row_set

	def get_column_nums(self, i):
		column_set = set()
		for x in col_indices(i, self.outer_square_size):
			if x != i:
				column_set.add(self.square[x])
		return column_set

	def get_inner_square_nums(self, i):
		inner_square_set = set()
		for x in subblock_indices(i, self.outer_square_size, self.inner_square_size):
			if x != i:
				inner_square_set.add(self.square[x])
		return inner_square_set

	def get_possible_nums(self, i):
		all = self.get_row_nums(i) | self.get_column_nums(i) | self.get_inner_square_nums(i)
		return self.sym_set - all

	def get_next_higher_possible_num(self, i, num):
		for pn in sorted(self.get_possible_nums(i)): # fixme: sorting is bad here
			if pn > num:
				return pn
		return self.sym_free

	def _solve(self):
		i = 0
		backtrack = []
		backtracks = 0
		steps = 0

		while i < self.outer_square_area:
			if self.square[i] != self.sym_free:
				i+=1
				continue
			pnums = self.get_possible_nums(i)
			#print(pnums)
			#self.print_square()

			if len(pnums) == 0:
				while True:
					backtracks += 1
					try:
						j, num = backtrack.pop()
					except IndexError:
						raise Unsolvable

					num = self.get_next_higher_possible_num(j, num)
					self.square[j] = num
					if num != self.sym_free:
						backtrack.append((j, num))
						i = j+1
						break
			else:
				num = min(pnums)
				self.square[i] = num
				backtrack.append((i, num))
				steps += 1

		return backtracks, steps

if __name__ == "__main__":
	from future.utils import viewitems
	from genutility.time import MeasureTime

	for name, sud in viewitems(sudokus):
		s = SudokuBruteforce(9, sud, set(range(1, 10)), 0)
		try:
			with MeasureTime() as t:
				steps, backtracks = s.solve()
		except Unsolvable:
			print("Cannot solve Sudoku")
		else:
			s.print_square()
		print("%s: Schritte: %u, Backtracks: %u, time: %f" % (name, steps, backtracks, t.get()))



from __future__ import absolute_import, division, print_function, unicode_literals

from functools import wraps
from warnings import warn

def deprecated(msg, stacklevel=2):
	def decorator(func):
		@wraps(func)
		def inner(*args, **kwargs):
			warn(msg, DeprecationWarning, stacklevel)
			print("DeprecationWarning:", msg)
			return func(*args, **kwargs)
		return inner
	return decorator
